from scifem import assemble_scalar

from festim.exports.surface_quantity import SurfaceQuantity


class AverageSurface(SurfaceQuantity):
    """Computes the average value of a field on a given surface

    Args:
        field (festim.Species): species for which the average surface is computed
        surface (festim.SurfaceSubdomain): surface subdomain
        filename (str, optional): name of the file to which the average surface is exported

    Attributes:
        see `festim.SurfaceQuantity`
    """

    @property
    def title(self):
        return f"Average {self.field.name} surface {self.surface.id}"

    def compute(self, ds):
        """
        Computes the average value of the field on the defined surface
        subdomain, and appends it to the data list
        """

        self.value = assemble_scalar(
            self.field.solution * ds(self.surface.id)
        ) / assemble_scalar(1 * ds(self.surface.id))
        self.data.append(self.value)



import csv

from festim.species import Species
from festim.subdomain.volume_subdomain import VolumeSubdomain


class VolumeQuantity:
    """Export VolumeQuantity

    Args:
        field: species for which the volume quantity is computed
        volume: volume subdomain
        filename: name of the file to which the volume quantity is exported

    Attributes:
        field: species for which the volume quantity is computed
        volume: volume subdomain
        filename: name of the file to which the volume quantity is exported
        t: list of time values
        data: list of values of the volume quantity
    """

    field: Species
    volume: VolumeSubdomain
    filename: str | None

    t: list[float]
    data: list[float]

    def __init__(self, field, volume, filename: str | None = None) -> None:
        self.field = field
        self.volume = volume
        self.filename = filename

        self.t = []
        self.data = []
        self._first_time_export = True

    @property
    def filename(self):
        return self._filename

    @filename.setter
    def filename(self, value):
        if value is None:
            self._filename = None
        elif not isinstance(value, str):
            raise TypeError("filename must be of type str")
        elif not value.endswith(".csv") and not value.endswith(".txt"):
            raise ValueError("filename must end with .csv or .txt")
        self._filename = value

    @property
    def field(self):
        return self._field

    @field.setter
    def field(self, value):
        # check that field is festim.Species
        if not isinstance(value, Species | str):
            raise TypeError("field must be of type festim.Species")

        self._field = value

    def write(self, t):
        """If the filename doesnt exist yet, create it and write the header,
        then append the time and value to the file"""

        if self.filename is not None:
            if self._first_time_export:
                header = ["t(s)", f"{self.title}"]
                with open(self.filename, mode="w+", newline="") as file:
                    writer = csv.writer(file)
                    writer.writerow(header)
                self._first_time_export = False
            with open(self.filename, mode="a", newline="") as file:
                writer = csv.writer(file)
                writer.writerow([t, self.value])



from unittest.mock import patch

import mpi4py.MPI as MPI

import dolfinx
import numpy as np

import festim as F


def test_writing_and_reading_of_species_function_using_checkpoints(tmpdir):
    """
    Tests that a model can write a checkpoint file and another model can read it.
    """
    mesh = dolfinx.mesh.create_unit_square(
        MPI.COMM_WORLD, nx=10, ny=10, cell_type=dolfinx.cpp.mesh.CellType.quadrilateral
    )
    my_model = F.HydrogenTransportProblem()
    my_model.mesh = F.Mesh(mesh)

    my_mat = F.Material(name="mat", D_0=1, E_D=0)
    vol = F.VolumeSubdomain(id=0, material=my_mat)
    surf = F.SurfaceSubdomain(id=1)
    my_model.subdomains = [vol, surf]

    H = F.Species("H")
    D = F.Species("D")
    my_model.species = [H, D]

    my_model.temperature = 500

    my_model.boundary_conditions = [
        F.FixedConcentrationBC(
            subdomain=surf, value=lambda x: 2 * x[0] + x[1], species=H
        ),
        F.FixedConcentrationBC(
            subdomain=surf, value=lambda x: 2 * x[1] + x[0], species=D
        ),
    ]

    my_model.settings = F.Settings(
        atol=1e-12, rtol=1e-12, transient=True, final_time=10
    )
    my_model.settings.stepsize = F.Stepsize(1)

    my_model.exports = [
        F.VTXSpeciesExport(
            filename=tmpdir + "/out_checkpoint.bp",
            field=[H, D],
            checkpoint=True,
        ),
        F.VTXSpeciesExport(
            filename=tmpdir + "/model_1_out_h.bp",
            field=[H],
        ),
    ]

    my_model.initialise()
    my_model.run()

    my_model2 = F.HydrogenTransportProblem()
    my_model2.mesh = F.Mesh(mesh)
    my_model2.subdomains = [vol, surf]

    H = F.Species("H")
    D = F.Species("D")
    my_model2.species = [H, D]

    my_model2.temperature = 500

    my_model2.initial_conditions = [
        F.InitialConcentration(
            value=F.read_function_from_file(
                filename=tmpdir + "/out_checkpoint.bp",
                name="H",
                timestamp=10,
            ),
            species=H,
            volume=vol,
        ),
        F.InitialConcentration(
            value=F.read_function_from_file(
                filename=tmpdir + "/out_checkpoint.bp",
                name="D",
                timestamp=10,
            ),
            species=D,
            volume=vol,
        ),
    ]

    my_model2.settings = F.Settings(
        atol=1e-10, rtol=1e-10, transient=True, final_time=10
    )
    my_model2.settings.stepsize = F.Stepsize(0.1)

    my_model2.exports = [
        F.VTXSpeciesExport(
            filename=tmpdir + "/model_2_out_h.bp",
            field=[H],
        ),
    ]

    my_model2.initialise()
    my_model2.run()

    import numpy as np

    np.testing.assert_allclose(
        H.post_processing_solution.x.array,
        1.5,
        atol=1e-10,
    )


def test_VTXExport_times_added_to_milestones(tmpdir):
    """Creates a HydrogenTransportProblem object and checks that, if no
    stepsize.milestones are given and VTXExport.times are given, VTXExport.times are
    are added to stepsize.milestones by .initialise()

    Args:
        tmpdir (os.PathLike): path to the pytest temporary folder
    """
    H = F.Species("H", mobile=True)
    filename = str(tmpdir.join("my_export.bp"))
    my_export = F.VTXSpeciesExport(
        field=H,
        filename=filename,
        times=[1, 2, 3],
    )

    # build
    my_model = F.HydrogenTransportProblem()
    my_model.mesh = F.Mesh1D([1, 2, 3])
    my_model.species = [H]
    my_model.subdomains = [
        F.VolumeSubdomain1D(id=1, borders=[1, 3], material=F.Material(D_0=1, E_D=0))
    ]
    my_model.temperature = 100
    my_model.settings = F.Settings(
        atol=0.999,
        rtol=0.999,
        final_time=4,
        transient=True,
        stepsize=F.Stepsize(initial_value=3),
    )
    my_model.exports = [my_export]

    # run
    my_model.initialise()

    # test
    assert my_model.settings.stepsize.milestones == my_export.times


def test_vtx_writer_called_only_at_specified_times(tmpdir):
    """test that the VTXWriter.write function is called the number of times specified in
    the export.times"""

    filename = str(tmpdir.join("my_export.bp"))

    my_model = F.HydrogenTransportProblem()

    my_model.mesh = F.Mesh1D(vertices=np.linspace(0, 10, 100))

    H = F.Species("H", mobile=True)
    my_model.species = [H]

    vol = F.VolumeSubdomain1D(
        id=1, borders=[0, 10], material=F.Material(D_0=1.0, E_D=0)
    )
    left = F.SurfaceSubdomain1D(id=2, x=0)
    my_model.subdomains = [vol, left]
    my_model.temperature = 500
    my_model.boundary_conditions = [
        F.FixedConcentrationBC(subdomain=left, value=5.0, species=H),
    ]
    my_model.exports = [F.VTXSpeciesExport(filename=filename, field=H, times=[2, 4, 6])]
    my_model.settings = F.Settings(
        atol=1e-10,
        rtol=1e-10,
        transient=True,
        final_time=6,
        stepsize=F.Stepsize(initial_value=1),
    )

    with patch("dolfinx.io.VTXWriter") as MockWriter:
        # This is the mock instance of VTXWriter
        mock_writer_instance = MockWriter.return_value

        my_model.initialise()
        my_model.run()

        # Check number of write calls
        assert mock_writer_instance.write.call_count == 3

        # Check which times were passed to write
        actual_times = [
            call.args[0] for call in mock_writer_instance.write.call_args_list
        ]
        expected_times = [2, 4, 6]
        assert actual_times == expected_times



from mpi4py import MPI

import dolfinx.mesh
import numpy as np
import pytest
import ufl
from dolfinx import fem

import festim as F


def sieverts_law(T, S_0, E_S, pressure):
    """Applies the Sieverts law to compute the concentration at the boundary"""
    S = S_0 * ufl.exp(-E_S / F.k_B / T)
    return S * pressure**0.5


def test_raise_error():
    """Test that a value error is raised if the pressure function is not supported in SievertsBC"""
    with pytest.raises(ValueError, match="pressure function not supported"):
        F.SievertsBC(
            subdomain=None, S_0=1.0, E_S=1.0, pressure=lambda c: c, species="H"
        )


@pytest.mark.parametrize(
    "pressure",
    [
        100,
        lambda x: x,
        lambda t: t,
        lambda T: T,
        lambda x, t: x + 2 * t,
        lambda x, T: x + 2 * T,
        lambda t, T: 2 * t + 3 * T,
        lambda x, t, T: x + 2 * t + 3 * T,
        lambda t: 1 if t < 0.5 else 2,
    ],
)
def test_create_new_value_function(pressure):
    my_BC = F.SievertsBC(
        subdomain=None, S_0=1.0, E_S=1.0, pressure=pressure, species="H"
    )
    assert my_BC.value is not None
    assert callable(my_BC.value)

    pressure_kwargs, value_kwargs = {}, {}
    if callable(pressure):
        if "x" in pressure.__code__.co_varnames:
            pressure_kwargs["x"] = 2.0
            value_kwargs["x"] = 2.0
        if "t" in pressure.__code__.co_varnames:
            pressure_kwargs["t"] = 3.0
            value_kwargs["t"] = 3.0
        if "T" in pressure.__code__.co_varnames:
            pressure_kwargs["T"] = 500.0
    value_kwargs["T"] = 500.0

    computed_value = my_BC.value(**value_kwargs)
    if callable(pressure):
        expected_value = sieverts_law(
            T=500.0, S_0=1.0, E_S=1.0, pressure=pressure(**pressure_kwargs)
        )
    else:
        expected_value = sieverts_law(T=500.0, S_0=1.0, E_S=1.0, pressure=pressure)
    assert np.isclose(computed_value, expected_value)


@pytest.mark.parametrize(
    "pressure",
    [
        100,
        lambda x: x[0],
        lambda t: t,
        lambda T: T,
        lambda x, t: x[0] + 2 * t,
        lambda x, T: x[0] + 2 * T,
        lambda t, T: 2 * t + 3 * T,
        lambda x, t, T: x[0] + 2 * t + 3 * T,
        lambda t: ufl.conditional(ufl.lt(t, 1.0), 100.0, 0.0),
        lambda t, x: ufl.conditional(ufl.lt(t, 1.0), 100.0 * x[0], 0.0),
    ],
)
def test_integration_with_HTransportProblem(pressure):
    subdomain = F.SurfaceSubdomain1D(1, x=1)
    dummy_mat = F.Material(1.0, 1.0, "dummy")
    vol_subdomain = F.VolumeSubdomain1D(1, borders=[0, 1], material=dummy_mat)

    mesh = dolfinx.mesh.create_unit_interval(MPI.COMM_WORLD, 10)
    my_model = F.HydrogenTransportProblem(
        mesh=F.Mesh(mesh), subdomains=[vol_subdomain, subdomain]
    )
    my_model.species = [F.Species("H")]
    my_bc = F.SievertsBC(
        subdomain=subdomain,
        S_0=2.0,
        E_S=0.5,
        pressure=pressure,
        species=my_model.species[0],
    )
    my_model.boundary_conditions = [my_bc]

    my_model.temperature = fem.Constant(my_model.mesh.mesh, 550.0)

    my_model.settings = F.Settings(atol=1, rtol=0.1, final_time=2)
    my_model.settings.stepsize = F.Stepsize(initial_value=1)

    # RUN

    my_model.initialise()

    assert my_bc.value_fenics is not None

    my_model.run()



from typing import Any

from mpi4py import MPI
from petsc4py import PETSc

import dolfinx
import numpy as np
import tqdm.autonotebook
import ufl
from dolfinx import fem
from dolfinx.nls.petsc import NewtonSolver
from packaging.version import Version

import festim as F
from festim.mesh.mesh import Mesh as _Mesh
from festim.source import SourceBase as _SourceBase
from festim.subdomain.volume_subdomain import (
    VolumeSubdomain as _VolumeSubdomain,
)


class ProblemBase:
    """
    Base class for :py:class:`HeatTransferProblem <festim.heat_transfer_problem.HeatTransferProblem>` and
    :py:class:`HydrogenTransportProblem <festim.hydrogen_transport_problem.HydrogenTransportProblem>`.

    Attributes:
        show_progress_bar: If `True` a progress bar is displayed during the simulation
        progress_bar: the progress bar
    """

    mesh: _Mesh
    sources: list[_SourceBase]
    exports: list[Any]
    subdomains: list[_VolumeSubdomain]
    show_progress_bar: bool
    progress_bar: None | tqdm.autonotebook.tqdm

    def __init__(
        self,
        mesh: _Mesh = None,
        sources=None,
        exports=None,
        subdomains=None,
        boundary_conditions=None,
        settings=None,
        petsc_options=None,
    ) -> None:
        self.mesh = mesh
        # for arguments to initialise as empty list
        # if arg not None, assign arg, else assign empty list
        self.subdomains = subdomains or []
        self.boundary_conditions = boundary_conditions or []
        self.sources = sources or []
        self.exports = exports or []
        self.settings = settings

        self.dx = None
        self.ds = None
        self.function_space = None
        self.facet_meshtags = None
        self.volume_meshtags = None
        self.formulation = None
        self.bc_forms = []
        self.show_progress_bar = True
        self.petsc_options = petsc_options

    @property
    def volume_subdomains(self):
        return [s for s in self.subdomains if isinstance(s, F.VolumeSubdomain)]

    @property
    def surface_subdomains(self):
        return [s for s in self.subdomains if isinstance(s, F.SurfaceSubdomain)]

    @property
    def dt(self):
        return self._dt

    def define_meshtags_and_measures(self):
        """Defines the facet and volume meshtags of the model which are used
        to define the measures fo the model, dx and ds"""

        if isinstance(self.mesh, F.MeshFromXDMF):
            # TODO: fix naming inconsistency between facet and surface meshtags
            self.facet_meshtags = self.mesh.define_surface_meshtags()
            self.volume_meshtags = self.mesh.define_volume_meshtags()

        elif (
            isinstance(self.mesh, F.Mesh)
            and self.facet_meshtags is None
            and self.volume_meshtags is None
        ):
            self.facet_meshtags, self.volume_meshtags = self.mesh.define_meshtags(
                surface_subdomains=self.surface_subdomains,
                volume_subdomains=self.volume_subdomains,
                # if self has attribute interfaces pass it
                interfaces=getattr(self, "interfaces", None),
            )

        # check volume ids are unique
        vol_ids = [vol.id for vol in self.volume_subdomains]
        if len(vol_ids) != len(np.unique(vol_ids)):
            raise ValueError("Volume ids are not unique")

        # define measures
        self.ds = ufl.Measure(
            "ds", domain=self.mesh.mesh, subdomain_data=self.facet_meshtags
        )
        self.dx = ufl.Measure(
            "dx", domain=self.mesh.mesh, subdomain_data=self.volume_meshtags
        )

    def define_boundary_conditions(self):
        """Defines the dirichlet boundary conditions of the model"""
        for bc in self.boundary_conditions:
            if isinstance(bc, F.DirichletBCBase):
                form = self.create_dirichletbc_form(bc)
                self.bc_forms.append(form)

    def create_solver(self):
        """Creates the solver of the model"""

        if Version(dolfinx.__version__) == Version("0.9.0"):
            problem = fem.petsc.NonlinearProblem(
                self.formulation,
                self.u,
                bcs=self.bc_forms,
            )
            self.solver = NewtonSolver(MPI.COMM_WORLD, problem)

            self.solver.atol = (
                self.settings.atol
                if not callable(self.settings.rtol)
                else self.settings.rtol(float(self.t))
            )
            self.solver.rtol = (
                self.settings.rtol
                if not callable(self.settings.rtol)
                else self.settings.rtol(float(self.t))
            )
            self.solver.max_it = self.settings.max_iterations

            ksp = self.solver.krylov_solver

            if self.petsc_options is None:
                ksp.setType("preonly")
                ksp.getPC().setType("lu")
                ksp.getPC().setFactorSolverType("mumps")
                ksp.setErrorIfNotConverged(True)
            else:
                # Set PETSc options
                opts = PETSc.Options()
                option_prefix = ksp.getOptionsPrefix()
                for k, v in self.petsc_options.items():
                    opts[f"{option_prefix}{k}"] = v
                ksp.setFromOptions()

        elif Version(dolfinx.__version__) > Version("0.9.0"):
            from dolfinx.fem.petsc import NonlinearProblem

            if self.petsc_options is None:
                # taken from https://github.com/FEniCS/dolfinx/blob/5fcb988c5b0f46b8f9183bc844d8f533a2130d6a/python/demo/demo_cahn-hilliard.py#L279C1-L286C28
                use_superlu = (
                    PETSc.IntType == np.int64
                )  # or PETSc.ScalarType == np.complex64
                sys = PETSc.Sys()  # type: ignore
                if sys.hasExternalPackage("mumps") and not use_superlu:
                    linear_solver = "mumps"
                elif sys.hasExternalPackage("superlu_dist"):
                    linear_solver = "superlu_dist"
                else:
                    linear_solver = "petsc"

                petsc_options = {
                    "snes_type": "newtonls",
                    "snes_linesearch_type": "none",
                    "snes_stol": np.sqrt(np.finfo(dolfinx.default_real_type).eps)
                    * 1e-2,
                    # TODO : make atol and rtol callable
                    "snes_atol": self.settings.atol,
                    "snes_rtol": self.settings.rtol,
                    "snes_max_it": self.settings.max_iterations,
                    "ksp_type": "preonly",
                    "pc_type": "lu",
                    "pc_factor_mat_solver_type": linear_solver,
                }
            else:
                petsc_options = self.petsc_options

            self.solver = NonlinearProblem(
                self.formulation,
                self.u,
                bcs=self.bc_forms,
                petsc_options=petsc_options,
                petsc_options_prefix="festim_solver",
            )
            # Delete PETSc options post setting them, ref:
            # https://gitlab.com/petsc/petsc/-/issues/1201
            snes = self.solver.solver
            prefix = snes.getOptionsPrefix()
            opts = PETSc.Options()
            for k in petsc_options.keys():
                del opts[f"{prefix}{k}"]

    def run(self):
        """Runs the model"""

        if self.settings.transient:
            # Solve transient
            if self.show_progress_bar:
                self.progress_bar = tqdm.autonotebook.tqdm(
                    desc=f"Solving {self.__class__.__name__}",
                    total=self.settings.final_time,
                    unit_scale=True,
                )
            while self.t.value < self.settings.final_time:
                self.iterate()
            if self.show_progress_bar:
                self.progress_bar.refresh()  # refresh progress bar to show 100%
                self.progress_bar.close()
        else:
            # Solve steady-state
            if Version(dolfinx.__version__) == Version("0.9.0"):
                self.solver.solve(self.u)
            elif Version(dolfinx.__version__) > Version("0.9.0"):
                self.solver.solve()
            self.post_processing()

    def iterate(self):
        """Iterates the model for a given time step"""
        if self.show_progress_bar:
            self.progress_bar.update(
                min(self.dt.value, abs(self.settings.final_time - self.t.value))
            )

        # update rtol if it's callable
        if callable(self.settings.rtol):
            self.solver.rtol = self.settings.rtol(self.t.value)
        # update rtol if it's callable
        if callable(self.settings.atol):
            self.solver.atol = self.settings.atol(self.t.value)

        self.t.value += self.dt.value

        self.update_time_dependent_values()

        # solve main problem
        if Version(dolfinx.__version__) == Version("0.9.0"):
            nb_its, converged = self.solver.solve(self.u)
        elif Version(dolfinx.__version__) > Version("0.9.0"):
            _ = self.solver.solve()
            converged_reason = self.solver.solver.getConvergedReason()
            assert converged_reason > 0
            nb_its = self.solver.solver.getIterationNumber()

        # post processing
        self.post_processing()

        # update previous solution
        self.u_n.x.array[:] = self.u.x.array[:]

        # adapt stepsize
        if self.settings.stepsize.adaptive:
            new_stepsize = self.settings.stepsize.modify_value(
                value=self.dt.value, nb_iterations=nb_its, t=self.t.value
            )
            self.dt.value = new_stepsize

    def update_time_dependent_values(self):
        t = float(self.t)
        for bc in self.boundary_conditions:
            if bc.time_dependent:
                bc.update(t=t)

        for source in self.sources:
            if source.value.explicit_time_dependent:
                source.value.update(t=t)



import mpi4py.MPI as MPI

import numpy as np
import ufl
from dolfinx import fem
from dolfinx.mesh import create_unit_cube, create_unit_square

from festim import Mesh1D

test_mesh_1d = Mesh1D(np.linspace(0, 1, 10000))
test_mesh_2d = create_unit_square(MPI.COMM_WORLD, 50, 50)
test_mesh_3d = create_unit_cube(MPI.COMM_WORLD, 20, 20, 20)
x_1d = ufl.SpatialCoordinate(test_mesh_1d.mesh)
x_2d = ufl.SpatialCoordinate(test_mesh_2d)
x_3d = ufl.SpatialCoordinate(test_mesh_3d)


def error_L2(u_computed, u_exact, degree_raise=3):
    # Create higher order function space
    degree = u_computed.function_space.ufl_element().degree
    family = u_computed.function_space.ufl_element().family_name
    mesh = u_computed.function_space.mesh
    W = fem.functionspace(mesh, (family, degree + degree_raise))
    # Interpolate approximate solution
    u_W = fem.Function(W)
    u_W.interpolate(u_computed)

    # Interpolate exact solution, special handling if exact solution
    # is a ufl expression or a python lambda function
    u_ex_W = fem.Function(W)
    if isinstance(u_exact, ufl.core.expr.Expr):
        u_expr = fem.Expression(u_exact, W.element.interpolation_points)
        u_ex_W.interpolate(u_expr)
    else:
        u_ex_W.interpolate(u_exact)

    # Compute the error in the higher order function space
    e_W = fem.Function(W)
    e_W.x.array[:] = u_W.x.array - u_ex_W.x.array

    # Integrate the error
    error = fem.form(ufl.inner(e_W, e_W) * ufl.dx)
    error_local = fem.assemble_scalar(error)
    error_global = mesh.comm.allreduce(error_local, op=MPI.SUM)
    return np.sqrt(error_global)



from pathlib import Path

import mpi4py

from dolfinx.io import XDMFFile

from festim.species import Species

from .vtx import ExportBaseClass


class XDMFExport(ExportBaseClass):
    """Export functions to XDMFfile

    Args:
        filename: The name of the output file
        field: The field(s) to export

    Attributes:
        _writer (dolfinx.io.XDMFFile): the XDMF writer
        _field (festim.Species, list of festim.Species): the field index to export
    """

    _mesh_written: bool
    _filename: Path
    _writer: XDMFFile | None

    def __init__(self, filename: str | Path, field: list[Species] | Species) -> None:
        # Initializes the writer
        self._writer = None
        super().__init__(filename, ".xdmf")
        self.field = field
        self._mesh_written = False

    @property
    def field(self) -> list[Species]:
        return self._field

    @field.setter
    def field(self, value: Species | list[Species]):
        # check that field is festim.Species or list of festim.Species
        if isinstance(value, list):
            for element in value:
                if not isinstance(element, Species):
                    raise TypeError(
                        f"Each element in the list must be a species, got {type(element)}."
                    )
            val = value
        elif isinstance(value, Species):
            val = [value]
        else:
            raise TypeError(
                f"field must be of type festim.Species or a list of festim.Species, got "
                f"{type(value)}."
            )
        self._field = val

    def define_writer(self, comm: mpi4py.MPI.Intracomm) -> None:
        """Define the writer

        Args:
            comm (mpi4py.MPI.Intracomm): the MPI communicator
        """
        self._writer = XDMFFile(comm, self.filename, "w")

    def write(self, t: float):
        """Write functions to VTX file

        Args:
            t (float): the time of export
        """
        if not self._mesh_written:
            self._writer.write_mesh(
                self.field[0].post_processing_solution.function_space.mesh
            )
            self._mesh_written = True

        for field in self.field:
            self._writer.write_function(field.post_processing_solution, t)

    def __del__(self):
        if self._writer is not None:
            self._writer.close()



import numpy as np

import festim as F

my_model = F.HydrogenTransportProblem()

# -------- Mesh --------- #

L = 20e-6
vertices = np.concatenate(
    [
        np.linspace(0, 30e-9, num=200, endpoint=False),
        np.linspace(30e-9, 3e-6, num=300, endpoint=False),
        np.linspace(3e-6, L, num=200),
    ]
)
my_model.mesh = F.Mesh1D(vertices)


# -------- Materials and subdomains --------- #

w_atom_density = 6.3e28  # atom/m3

tungsten = F.Material(D_0=4.1e-7, E_D=0.39, name="tungsten")

my_subdomain = F.VolumeSubdomain1D(id=1, borders=[0, L], material=tungsten)
left_surface = F.SurfaceSubdomain1D(id=1, x=0)
right_surface = F.SurfaceSubdomain1D(id=2, x=L)

my_model.subdomains = [
    my_subdomain,
    left_surface,
    right_surface,
]

# -------- Hydrogen species and reactions --------- #

mobile_H = F.Species("H")
trapped_H1 = F.Species("trapped_H1", mobile=False)
empty_trap1 = F.ImplicitSpecies(
    n=1.3e-3 * w_atom_density, others=[trapped_H1], name="empty_trap1"
)
trapped_H2 = F.Species("trapped_H2", mobile=False)
empty_trap2 = F.ImplicitSpecies(
    n=4e-3 * w_atom_density, others=[trapped_H2], name="empty_trap2"
)
my_model.species = [mobile_H, trapped_H1, trapped_H2]

my_model.reactions = [
    F.Reaction(
        k_0=4.1e-7 / (1.1e-10**2 * 6 * w_atom_density),
        E_k=0.39,
        p_0=1e13,
        E_p=0.87,
        reactant=[mobile_H, empty_trap1],
        product=trapped_H1,
        volume=my_subdomain,
    ),
    F.Reaction(
        k_0=4.1e-7 / (1.1e-10**2 * 6 * w_atom_density),
        E_k=0.39,
        p_0=1e13,
        E_p=1.0,
        reactant=[mobile_H, empty_trap2],
        product=trapped_H2,
        volume=my_subdomain,
    ),
]

# -------- Temperature --------- #

implantation_time = 400
start_tds = implantation_time + 50
implantation_temp = 300
temperature_ramp = 8  # K/s


def temp_function(t):
    if t < start_tds:
        return implantation_temp
    else:
        return implantation_temp + temperature_ramp * (t - start_tds)


my_model.temperature = temp_function

# -------- Boundary conditions --------- #


def left_conc_value(t):
    if t < implantation_time:
        _T = temp_function(t)
        D = tungsten.D_0 * np.exp(-tungsten.E_D / F.k_B / _T)
        return 2.5e19 * 4.5e-9 / D
    else:
        return 0.0


my_model.boundary_conditions = [
    F.DirichletBC(subdomain=left_surface, value=left_conc_value, species=mobile_H),
    F.DirichletBC(subdomain=right_surface, value=0, species=mobile_H),
]

# -------- Exports --------- #

left_flux = F.SurfaceFlux(field=mobile_H, surface=left_surface)
right_flux = F.SurfaceFlux(field=mobile_H, surface=right_surface)

my_model.exports = [
    # F.VTXSpeciesExport("mobile_concentration_h.bp", field=mobile_H),
    # F.VTXSpeciesExport("trapped_concentration_h.bp", field=trapped_H1),
    F.XDMFExport("mobile_concentration_h.xdmf", field=mobile_H),
    F.XDMFExport("trapped_concentration_h1.xdmf", field=trapped_H1),
    F.XDMFExport("trapped_concentration_h2.xdmf", field=trapped_H2),
    left_flux,
    right_flux,
]

# -------- Settings --------- #

my_model.settings = F.Settings(atol=1e10, rtol=1e-10, max_iterations=30, final_time=500)

my_model.settings.stepsize = F.Stepsize(initial_value=0.5)

# -------- Run --------- #

my_model.initialise()
my_model.run()

# -------- Save results --------- #

np.savetxt(
    "outgassing_flux_tds.txt",
    np.array(left_flux.data) + np.array(right_flux.data),
)
np.savetxt("times_tds.txt", np.array(left_flux.t))



import numpy as np
import ufl
from dolfinx import fem

import festim as F

dummy_mat = F.Material(D_0=1, E_D=1, name="dummy")


def test_average_volume_compute_1D():
    """Test that the average volume export computes the correct value"""

    # BUILD
    L = 6.0
    my_mesh = F.Mesh1D(np.linspace(0, L, 10000))
    dummy_volume = F.VolumeSubdomain1D(id=1, borders=[0, L], material=dummy_mat)
    temp, cell_meshtags = my_mesh.define_meshtags(
        surface_subdomains=[], volume_subdomains=[dummy_volume]
    )
    dx = ufl.Measure("dx", domain=my_mesh.mesh, subdomain_data=cell_meshtags)

    # give function to species
    V = fem.functionspace(my_mesh.mesh, ("Lagrange", 1))
    c = fem.Function(V)
    c.interpolate(lambda x: x[0] * 0.5 + 2)

    my_species = F.Species("H")
    my_species.solution = c

    my_export = F.AverageVolume(field=my_species, volume=dummy_volume)

    # RUN
    my_export.compute(my_species.solution, dx)

    # TEST
    expected_value = 3.5
    computed_value = my_export.value

    assert np.isclose(computed_value, expected_value, rtol=1e-2)



import numpy as np
import pytest

import festim as F


@pytest.mark.parametrize("times", [None, [2, 5], [5]])
def test_profile(times):
    my_model = F.HydrogenTransportProblem()

    protium = F.Species("H")
    deuterium = F.Species("D")
    tritium = F.Species("T")
    my_model.species = [protium, deuterium, tritium]

    my_model.mesh = F.Mesh1D(np.linspace(0, 1, 100))

    left_surf = F.SurfaceSubdomain1D(id=1, x=0)
    right_surf = F.SurfaceSubdomain1D(id=2, x=1)

    # assumes the same diffusivity for all species
    material = F.Material(D_0=1, E_D=0)

    vol = F.VolumeSubdomain1D(id=1, borders=[0, 1], material=material)

    my_model.subdomains = [vol, left_surf, right_surf]

    my_model.boundary_conditions = [
        # Protium BCs
        F.FixedConcentrationBC(left_surf, value=10, species=protium),
        F.FixedConcentrationBC(right_surf, value=0, species=protium),
        # Deuterium BCs
        F.FixedConcentrationBC(left_surf, value=5, species=deuterium),
        F.FixedConcentrationBC(right_surf, value=0, species=deuterium),
        # Tritium BCs
        F.FixedConcentrationBC(left_surf, value=0, species=tritium),
        F.FixedConcentrationBC(right_surf, value=2, species=tritium),
    ]

    my_model.temperature = 300

    my_model.settings = F.Settings(atol=1e-10, rtol=1e-10, final_time=5)

    my_model.settings.stepsize = F.Stepsize(1)

    my_model.exports = [
        F.Profile1DExport(protium, times=times),
        F.Profile1DExport(deuterium, times=times),
    ]

    my_model.initialise()
    my_model.run()

    assert my_model.exports[0].x is not None
    assert my_model.exports[1].x is not None
    if times is None:
        assert len(my_model.exports[0].data) > 0
        assert len(my_model.exports[1].data) > 0
    else:
        assert len(my_model.exports[0].data) == len(times)
        assert len(my_model.exports[1].data) == len(times)


def test_profile_single_species():
    my_model = F.HydrogenTransportProblem()

    protium = F.Species("H")
    my_model.species = [protium]

    my_model.mesh = F.Mesh1D(np.linspace(0, 1, 100))

    left_surf = F.SurfaceSubdomain1D(id=1, x=0)
    right_surf = F.SurfaceSubdomain1D(id=2, x=1)

    # assumes the same diffusivity for all species
    material = F.Material(D_0=1, E_D=0)

    my_model.species = [protium]

    my_model.mesh = F.Mesh1D(np.linspace(0, 1, 100))

    left_surf = F.SurfaceSubdomain1D(id=1, x=0)
    right_surf = F.SurfaceSubdomain1D(id=2, x=1)

    # assumes the same diffusivity for all species
    material = F.Material(D_0=1, E_D=0)

    vol = F.VolumeSubdomain1D(id=1, borders=[0, 1], material=material)

    my_model.subdomains = [vol, left_surf, right_surf]

    my_model.boundary_conditions = [
        F.FixedConcentrationBC(left_surf, value=10, species=protium),
        F.FixedConcentrationBC(right_surf, value=0, species=protium),
    ]

    my_model.temperature = 300

    my_model.settings = F.Settings(atol=1e-10, rtol=1e-10, final_time=5)

    my_model.settings.stepsize = F.Stepsize(1)

    my_model.exports = [
        F.Profile1DExport(protium),
    ]

    my_model.initialise()
    my_model.run()

    assert my_model.exports[0].x is not None
    assert len(my_model.exports[0].data) > 0


def test_profile_discontinuous():
    my_model = F.HydrogenTransportProblemDiscontinuous()

    protium = F.Species("H")
    deuterium = F.Species("D")
    tritium = F.Species("T")
    my_model.species = [protium, deuterium, tritium]

    vertices_left = np.linspace(0, 0.5, 50)
    vertices_right = np.linspace(0.5, 1, 50)
    vertices = np.concatenate((vertices_left, vertices_right))

    my_model.mesh = F.Mesh1D(vertices)

    left_surf = F.SurfaceSubdomain1D(id=1, x=0)
    right_surf = F.SurfaceSubdomain1D(id=2, x=1)

    # assumes the same diffusivity for all species
    material_left = F.Material(D_0=1, E_D=0, K_S_0=1, E_K_S=0)
    material_right = F.Material(D_0=1, E_D=0, K_S_0=2, E_K_S=0)

    vol1 = F.VolumeSubdomain1D(id=1, borders=[0, 0.5], material=material_left)
    vol2 = F.VolumeSubdomain1D(id=2, borders=[0.5, 1], material=material_right)

    my_model.interfaces = [
        F.Interface(id=3, subdomains=[vol1, vol2], penalty_term=1000)
    ]

    my_model.subdomains = [vol1, vol2, left_surf, right_surf]

    for spe in my_model.species:
        spe.subdomains = [vol1, vol2]

    my_model.surface_to_volume = {
        left_surf: vol1,
        right_surf: vol2,
    }

    my_model.boundary_conditions = [
        # Protium BCs
        F.FixedConcentrationBC(left_surf, value=10, species=protium),
        F.FixedConcentrationBC(right_surf, value=0, species=protium),
        # Deuterium BCs
        F.FixedConcentrationBC(left_surf, value=5, species=deuterium),
        F.FixedConcentrationBC(right_surf, value=0, species=deuterium),
        # Tritium BCs
        F.FixedConcentrationBC(left_surf, value=0, species=tritium),
        F.FixedConcentrationBC(right_surf, value=2, species=tritium),
    ]

    my_model.temperature = 300

    my_model.settings = F.Settings(atol=1e-10, rtol=1e-10, final_time=0.5)

    my_model.settings.stepsize = F.Stepsize(0.1)

    my_model.exports = [
        F.Profile1DExport(field=protium, subdomain=vol1),
        F.Profile1DExport(field=deuterium, subdomain=vol1),
        F.Profile1DExport(field=tritium, subdomain=vol1),
        F.Profile1DExport(field=protium, subdomain=vol2),
        F.Profile1DExport(field=deuterium, subdomain=vol2),
        F.Profile1DExport(field=tritium, subdomain=vol2),
    ]

    my_model.initialise()
    my_model.run()

    for export in my_model.exports:
        assert export.x is not None
        assert len(export.data) > 0


def test_profile_discontinuous_single_species():
    my_model = F.HydrogenTransportProblemDiscontinuous()

    protium = F.Species("H")
    my_model.species = [protium]

    vertices_left = np.linspace(0, 0.5, 50)
    vertices_right = np.linspace(0.5, 1, 50)
    vertices = np.concatenate((vertices_left, vertices_right))

    my_model.mesh = F.Mesh1D(vertices)

    left_surf = F.SurfaceSubdomain1D(id=1, x=0)
    right_surf = F.SurfaceSubdomain1D(id=2, x=1)

    # assumes the same diffusivity for all species
    material_left = F.Material(D_0=1, E_D=0, K_S_0=1, E_K_S=0)
    material_right = F.Material(D_0=1, E_D=0, K_S_0=2, E_K_S=0)

    vol1 = F.VolumeSubdomain1D(id=1, borders=[0, 0.5], material=material_left)
    vol2 = F.VolumeSubdomain1D(id=2, borders=[0.5, 1], material=material_right)

    my_model.interfaces = [
        F.Interface(id=3, subdomains=[vol1, vol2], penalty_term=1000)
    ]

    my_model.subdomains = [vol1, vol2, left_surf, right_surf]

    for spe in my_model.species:
        spe.subdomains = [vol1, vol2]

    my_model.surface_to_volume = {
        left_surf: vol1,
        right_surf: vol2,
    }

    my_model.boundary_conditions = [
        F.FixedConcentrationBC(left_surf, value=10, species=protium),
        F.FixedConcentrationBC(right_surf, value=0, species=protium),
    ]

    my_model.temperature = 300

    my_model.settings = F.Settings(atol=1e-10, rtol=1e-10, final_time=0.5)

    my_model.settings.stepsize = F.Stepsize(0.1)

    my_model.exports = [
        F.Profile1DExport(field=protium, subdomain=vol1),
        F.Profile1DExport(field=protium, subdomain=vol2),
    ]

    my_model.initialise()
    my_model.run()

    for export in my_model.exports:
        assert export.x is not None
        assert len(export.data) > 0



from collections.abc import Callable
from typing import Union

from mpi4py import MPI

import adios4dolfinx
import dolfinx
import numpy as np
import ufl
from dolfinx import fem

from festim.helpers import get_interpolation_points
from festim.species import Species
from festim.subdomain.volume_subdomain import VolumeSubdomain


class InitialConditionBase:
    """
    Base initial condition class

    Args:
        value: the value of the initial condition.
        volume: the volume subdomain where the initial condition is applied

    Attributes:
        value: the value of the initial condition.
        volume: the volume subdomain where the initial condition is applied
    """

    value: (
        float
        | int
        | fem.Constant
        | np.ndarray
        | fem.Expression
        | ufl.core.expr.Expr
        | fem.Function
    )
    volume: VolumeSubdomain

    def __init__(
        self,
        value: (
            float
            | int
            | fem.Constant
            | np.ndarray
            | fem.Expression
            | ufl.core.expr.Expr
            | fem.Function
        ),
        volume: VolumeSubdomain,
    ):
        self.value = value
        self.volume = volume

    @property
    def volume(self):
        return self._volume

    @volume.setter
    def volume(self, value):
        # check that volume is festim.VolumeSubdomain
        if not isinstance(value, VolumeSubdomain):
            raise TypeError("volume must be of type festim.VolumeSubdomain")
        self._volume = value


class InitialConcentration(InitialConditionBase):
    """
    Initial concentration class

    Args:
        value: the value of the initial concentration of a given species.
        species: the species to which the condition is applied
        volume: the volume subdomain where the initial condition is applied

    Attributes:
        value: the value of the initial concentration of a given species.
        species: the species to which the condition is applied
        volume: the volume subdomain where the initial condition is applied
        expr_fenics: the value of the initial condition in fenics expr format

    Examples:

        .. testsetup:: InitialConcentration

            from festim import InitialConcentration, Species, Material, VolumeSubdomain
            my_species = Species(name='test')
            dummy_mat = Material(D_0=1, E_D=0.1)
            my_vol = VolumeSubdomain(id=1, material=dummy_mat)

        .. testcode:: InitialConcentration

            InitialConcentration(value=1, species=my_species, volume=my_vol)
            InitialConcentration(
                value=lambda x: 1 + x[0],
                species=my_species,
                volume=my_vol
            )
            InitialConcentration(
                value=lambda T: 1 + T,
                species=my_species,
                volume=my_vol
            )
            InitialConcentration(
                value=lambda x, T: 1 + x[0] + T,
                species=my_species,
                volume=my_vol
            )
    """

    expr_fenics: Union[Callable, fem.Expression]
    species: Species

    def __init__(self, value, volume, species: Species):
        super().__init__(value=value, volume=volume)

        self.species = species

        self.expr_fenics = None

    @property
    def species(self):
        return self._species

    @species.setter
    def species(self, value):
        # check that species is festim.Species or list of festim.Species
        if not isinstance(value, Species):
            raise TypeError("species must be of type festim.Species")

        self._species = value

    def create_expr_fenics(
        self,
        mesh: dolfinx.mesh.Mesh,
        temperature: fem.Function | fem.Constant,
        function_space: fem.functionspace,
    ):
        """Creates the expr_fenics of the initial condition.

        If the value is a float or int, a function is created with an array with the
        shape of the mesh and all set to the value. Otherwise, it is converted to a
        fem.Expression.

        Args:
            mesh: the mesh
            temperature: the temperature
            function_space: the function space of the species
        """
        x = ufl.SpatialCoordinate(mesh)

        if isinstance(self.value, int | float):
            self.expr_fenics = lambda x: np.full(x.shape[1], self.value)
        elif isinstance(self.value, fem.Function):
            self.expr_fenics = self.value
        elif callable(self.value):
            arguments = self.value.__code__.co_varnames
            kwargs = {}
            if "t" in arguments:
                raise ValueError("Initial condition cannot be a function of time.")
            if "x" in arguments:
                kwargs["x"] = x
            if "T" in arguments:
                kwargs["T"] = temperature

            self.expr_fenics = fem.Expression(
                self.value(**kwargs),
                get_interpolation_points(function_space.element),
            )


class InitialTemperature(InitialConditionBase):
    """
    Initial temperature class

    Args:
        value: the value of the initial temperature
        volume: the volume subdomain where the initial condition is applied

    Attributes:
        value: the value of the initial temperature
        volume: the volume subdomain where the initial condition is applied
        expr_fenics: the value of the initial condition in fenics expr format

    Examples:

        .. testsetup:: InitialTemperature

            from festim import InitialTemperature, Material, VolumeSubdomain
            dummy_mat = Material(D_0=1, E_D=0.1)
            my_vol = VolumeSubdomain(id=1, material=dummy_mat)

        .. testcode:: InitialTemperature

            InitialTemperature(value=1, volume=my_vol)
            InitialTemperature(value=lambda x: 1 + x[0], volume=my_vol)
            InitialTemperature(value=lambda x, t: 1 + x[0] + t, volume=my_vol)
    """

    def __init__(self, value, volume):
        super().__init__(value=value, volume=volume)

        self.expr_fenics = None

    def create_expr_fenics(
        self,
        mesh: dolfinx.mesh.Mesh,
        function_space: fem.functionspace,
    ):
        """Creates the expr_fenics of the initial condition.

        If the value is a float or int, a function is created with an array with the
        shape of the mesh and all set to the value. Otherwise, it is converted to a
        fem.Expression.

        Args:
            mesh: the mesh
            function_space: the function space of the species
        """
        x = ufl.SpatialCoordinate(mesh)

        if isinstance(self.value, int | float):
            self.expr_fenics = lambda x: np.full(x.shape[1], self.value)
        elif isinstance(self.value, fem.Function):
            self.expr_fenics = self.value
        elif callable(self.value):
            arguments = self.value.__code__.co_varnames
            kwargs = {}
            if "t" in arguments:
                raise ValueError("Initial condition cannot be a function of time.")
            if "x" in arguments:
                kwargs["x"] = x

            self.expr_fenics = fem.Expression(
                self.value(**kwargs),
                get_interpolation_points(function_space.element),
            )


def read_function_from_file(
    filename: str, name: str, timestamp: int | float, family="P", order: int = 1
) -> fem.Function:
    """
    Read a function from a file

    note::
        The function is read from a file using adios4dolfinx. For more information
        see the [adios4dolfinx documentation](https://jsdokken.com/adios4dolfinx/README.html).

    Args:
        filename: the filename
        name: the name of the function
        timestamp: the timestamp of the function
        family: the family of the function space
        order: the order of the function space

    Returns:
        the function
    """
    mesh_in = adios4dolfinx.read_mesh(filename, MPI.COMM_WORLD)
    V_in = fem.functionspace(mesh_in, (family, order))
    u_in = fem.Function(V_in)
    adios4dolfinx.read_function(
        filename=filename,
        u=u_in,
        name=name,
        time=timestamp,
    )
    return u_in



import pytest

import festim as F

dummy_mat = F.Material(D_0=1, E_D=1, name="dummy")
dummy_volume = F.VolumeSubdomain1D(id=1, borders=[0, 1], material=dummy_mat)


def test_filename_setter_raises_TypeError():
    """Test that a TypeError is raised when the filename is not a string"""

    with pytest.raises(TypeError, match="filename must be of type str"):
        F.VolumeQuantity(
            filename=1,
            field=F.Species("test"),
            volume=dummy_volume,
        )


def test_filename_setter_raises_ValueError():
    """Test that a ValueError is raised when the filename does not end with .csv or .txt"""

    with pytest.raises(ValueError):
        F.VolumeQuantity(
            filename="my_export.xdmf",
            field=F.Species("test"),
            volume=dummy_volume,
        )


def test_field_setter_raises_TypeError():
    """Test that a TypeError is raised when the field is not a F.Species"""

    with pytest.raises(TypeError):
        F.VolumeQuantity(
            field=1,
            volume=dummy_volume,
        )



from collections.abc import Callable

import dolfinx
import numpy as np


class SurfaceSubdomain:
    """
    Surface subdomain class

    Args:
        id: the id of the surface subdomain
        locator: a callable function that locates the boundary facets of the subdomain

    Examples:

        .. testsetup:: SurfaceSubdomain

            from festim import SurfaceSubdomain

        .. testcode:: SurfaceSubdomain

            SurfaceSubdomain(id=1, locator=lambda x: np.isclose(x[0], 1.0))
            SurfaceSubdomain(id=1, locator=lambda x: np.logical_or(np.isclose(x[1], 0.0), np.isclose(x[1], 1.0)))
            SurfaceSubdomain(id=1, locator=lambda x: np.logical_and(np.isclose(x[0], 0.0), np.isclose(x[1], 1.0)))
            SurfaceSubdomain(id=1, locator=lambda x: np.logical_and(np.isclose(x[0], 0.0), x[1] <= 0.5))
    """

    id: int
    locator: Callable

    def __init__(self, id: int, locator: Callable | None = None):
        self.id = id
        self.locator = locator

    def locate_boundary_facet_indices(self, mesh: dolfinx.mesh.Mesh) -> np.ndarray:
        """Locate boundary facets of the subdomain in the mesh.

        Args:
            mesh: a dolfinx mesh object

        Raises:
            ValueError: if no locator function is provided

        Returns:
            the list of entities (facets) that belong to the subdomain
        """
        if self.locator is None:
            raise ValueError(
                "No locator function provided for locating boundary facets."
            )

        return dolfinx.mesh.locate_entities_boundary(
            mesh, mesh.topology.dim - 1, self.locator
        )


class SurfaceSubdomain1D(SurfaceSubdomain):
    """
    Surface subdomain class for 1D cases

    Args:
        id: the id of the surface subdomain
        x: the x coordinate of the surface subdomain

    Attributes:
        id: the id of the surface subdomain
        x: the x coordinate of the surface subdomain

    Examples:

        .. testsetup:: SurfaceSubdomain1D

            from festim import SurfaceSubdomain1D

        .. testcode:: SurfaceSubdomain1D

            SurfaceSubdomain1D(id=1, x=1)
    """

    # FIXME: Rename this to _id and use getter/setter
    id: int
    x: float

    def __init__(self, id: int, x: float) -> None:
        super().__init__(id, locator=lambda x_: np.isclose(x_[0], x))
        self.x = x


def find_surface_from_id(id: int, surfaces: list):
    """Returns the correct surface subdomain object from a list of surface ids
    based on an int

    Args:
        id (int): the id of the surface subdomain
        surfaces (list of F.SurfaceSubdomain): the list of surfaces

    Returns:
        festim.SurfaceSubdomain: the surface subdomain object with the correct id

    Raises:
        ValueError: if the surface name is not found in the list of surfaces

    """
    for surf in surfaces:
        if surf.id == id:
            return surf
    raise ValueError(f"id {id} not found in list of surfaces")



from mpi4py import MPI

import dolfinx.mesh
import pytest
import ufl
from dolfinx import fem

import festim as F

dummy_mat = F.Material(D_0=1, E_D=1, name="dummy_mat")

mesh = dolfinx.mesh.create_unit_interval(MPI.COMM_WORLD, 10)


def test_init():
    """Test that the attributes are set correctly"""
    # create a Source object
    volume = F.VolumeSubdomain1D(id=1, borders=[0, 1], material=dummy_mat)
    value = 1.0
    species = F.Species("test")
    source = F.ParticleSource(volume=volume, value=value, species=species)

    # check that the attributes are set correctly
    assert source.volume == volume
    assert source.species == species

    # check value is processed correctly
    assert source.value.input_value == value
    assert isinstance(source.value, F.Value)


@pytest.mark.parametrize(
    "value, expected_type",
    [
        (1.0, fem.Constant),
        (1, fem.Constant),
        (lambda t: t, fem.Constant),
        (lambda t: 1.0 + t, fem.Constant),
        (lambda x: 1.0 + x[0], ufl.core.expr.Expr),
        (lambda x, t: 1.0 + x[0] + t, ufl.core.expr.Expr),
        (lambda x, t, T: 1.0 + x[0] + t + T, ufl.core.expr.Expr),
        (
            lambda x, t: ufl.conditional(ufl.lt(t, 1.0), 100.0 + x[0], 0.0),
            ufl.core.expr.Expr,
        ),
    ],
)
def test_create_fenics_object(value, expected_type):
    """Test that the correct fenics object is created depending on the value input"""

    # BUILD
    vol_subdomain = F.VolumeSubdomain1D(1, borders=[0, 1], material=dummy_mat)
    species = F.Species("test")
    V = fem.functionspace(mesh, ("Lagrange", 1))

    source = F.ParticleSource(volume=vol_subdomain, value=value, species=species)

    T = fem.Constant(mesh, 550.0)
    t = fem.Constant(mesh, 0.0)

    # RUN
    source.value.convert_input_value(
        function_space=V, temperature=T, t=t, up_to_ufl_expr=True
    )

    # TEST
    # check that the value_fenics attribute is set correctly
    assert isinstance(source.value.fenics_object, expected_type)


@pytest.mark.parametrize(
    "input, expected_value",
    [
        (1.0, False),
        (None, False),
        (fem.Constant(mesh, dolfinx.default_scalar_type(1.0)), False),
        (lambda t: t, True),
        (lambda t: 1.0 + t, True),
        (lambda x: 1.0 + x[0], False),
        (lambda x, t: 1.0 + x[0] + t, True),
        (lambda x, t, T: 1.0 + x[0] + t + T, True),
        (lambda x, t: ufl.conditional(ufl.lt(t, 1.0), 100.0 + x[0], 0.0), True),
    ],
)
def test_source_explicit_time_dependent_attribute(input, expected_value):
    """Test that the time_dependent attribute is correctly set"""
    volume = F.VolumeSubdomain1D(1, borders=[0, 1], material=dummy_mat)
    species = F.Species("test")
    my_source = F.ParticleSource(input, volume, species)

    assert my_source.value.explicit_time_dependent is expected_value


@pytest.mark.parametrize(
    "input, expected_value",
    [
        (1.0, False),
        (None, False),
        (fem.Constant(mesh, dolfinx.default_scalar_type(1.0)), False),
        (lambda T: T, True),
        (lambda t: 1.0 + t, False),
        (lambda x, T: 1.0 + x[0] + T, True),
        (lambda x, t, T: 1.0 + x[0] + t + T, True),
        (
            lambda x, t: ufl.conditional(ufl.lt(t, 1.0), 100.0 + x[0], 0.0),
            False,
        ),
    ],
)
def test_source_temperature_dependent_attribute(input, expected_value):
    """Test that the temperature_dependent attribute is correctly set"""
    volume = F.VolumeSubdomain1D(1, borders=[0, 1], material=dummy_mat)
    species = F.Species("test")
    my_source = F.ParticleSource(input, volume, species)

    assert my_source.value.temperature_dependent is expected_value


def test_ValueError_raised_when_callable_returns_wrong_type():
    """The create_value method should raise a ValueError when the callable
    returns an object which is not a float or int"""

    vol_subdomain = F.VolumeSubdomain1D(1, borders=[0, 1], material=dummy_mat)
    species = F.Species("test")

    V = fem.functionspace(mesh, ("Lagrange", 1))

    def my_value(t):
        return ufl.conditional(ufl.lt(t, 0.5), 100, 0)

    source = F.ParticleSource(volume=vol_subdomain, value=my_value, species=species)

    T = fem.Constant(mesh, 550.0)
    t = fem.Constant(mesh, 0.0)

    with pytest.raises(
        ValueError,
        match="self.value should return a float or an int, not <class 'ufl.conditional.Conditional'",
    ):
        source.value.convert_input_value(
            function_space=V, temperature=T, t=t, up_to_ufl_expr=True
        )


def test_ValueError_raised_when_callable_returns_wrong_type_heat_source():
    """The create_value method should raise a ValueError when the callable
    returns an object which is not a float or int"""

    vol_subdomain = F.VolumeSubdomain1D(1, borders=[0, 1], material=dummy_mat)

    def my_value(t):
        return ufl.conditional(ufl.lt(t, 0.5), 100, 0)

    source = F.HeatSource(volume=vol_subdomain, value=my_value)

    t = fem.Constant(mesh, 0.0)

    V = fem.functionspace(mesh, ("Lagrange", 1))

    with pytest.raises(
        ValueError,
        match="self.value should return a float or an int, not <class 'ufl.conditional.Conditional'",
    ):
        source.value.convert_input_value(function_space=V, t=t, up_to_ufl_expr=True)


@pytest.mark.parametrize(
    "volume_input",
    [
        1.0,
        "1",
        ["1"],
        [1.0],
        [[1]],
        [[F.VolumeSubdomain1D(1, borders=[0, 1], material=dummy_mat)]],
        [[F.VolumeSubdomain(1, material=dummy_mat)]],
        None,
    ],
)
def test_TypeError_is_raised_when_volume_wrong_type(volume_input):
    """Test that a TypeError is raised when the volume is not of type
    festim.VolumeSubdomain"""

    my_spe = F.Species("test")
    with pytest.raises(
        TypeError,
        match="volume must be of type festim.VolumeSubdomain",
    ):
        F.ParticleSource(volume=volume_input, value=1.0, species=my_spe)


@pytest.mark.parametrize(
    "species_input",
    [
        1,
        1.0,
        [1],
        [1.0],
        [["test"]],
        [[F.Species("test")]],
        None,
    ],
)
def test_TypeError_is_raised_when_species_wrong_type(species_input):
    """Test that a TypeError is raised when the species is not of type
    festim.Species"""

    my_vol = F.VolumeSubdomain1D(id=1, borders=[0, 1], material=dummy_mat)

    with pytest.raises(
        TypeError,
        match="species must be of type festim.Species",
    ):
        F.ParticleSource(volume=my_vol, value=1.0, species=species_input)



import numpy as np
import pytest

import festim as F


@pytest.mark.parametrize("test_type", [int, F.Stepsize, float])
def test_stepsize_value(test_type):
    """Test that the stepsize is correctly set"""
    test_value = 23.0
    my_settings = F.Settings(atol=1, rtol=0.1)
    my_settings.stepsize = test_type(test_value)

    assert isinstance(my_settings.stepsize, F.Stepsize)
    assert np.isclose(my_settings.stepsize.initial_value, test_value)


def test_stepsize_value_wrong_type():
    """Checks that an error is raised when the wrong type is given"""
    my_settings = F.Settings(atol=1, rtol=0.1)

    with pytest.raises(TypeError):
        my_settings.stepsize = "coucou"


# @pytest.mark.parametrize(
#     "rtol",
#     [1e-10, lambda t: 1e-8 if t < 10 else 1e-10],
# )
# def test_callable_rtol(rtol):
#     """Tests callable rtol."""
#     my_settings = F.Settings(atol=0.1, rtol=rtol)

#     assert my_settings.rtol == rtol


# @pytest.mark.parametrize("atol", [1e10, lambda t: 1e12 if t < 10 else 1e10])
# def test_callable_atol(atol):
#     """Tests callable atol."""
#     my_settings = F.Settings(atol=atol, rtol=0.1)

#     assert my_settings.atol == atol


# @pytest.mark.parametrize(
#     "rtol, atol",
#     [
#         (lambda t: 1e-8 if t < 10 else 1e-10, lambda t: 1e12 if t < 10 else 1e10),
#     ],
# )
# def test_tolerances_value(rtol, atol):
#     """Tests that callable tolerances are called & return correct float before passed to fenics"""

#     # BUILD
#     test_mesh = F.Mesh1D(vertices=np.array([0.0, 1.0, 2.0, 3.0, 4.0]))
#     dummy_mat = F.Material(D_0=1, E_D=1, name="dummy_mat")

#     my_vol = F.VolumeSubdomain1D(id=1, borders=[0, 4], material=dummy_mat)
#     my_model = F.HydrogenTransportProblem(
#         mesh=test_mesh,
#         settings=F.Settings(atol=atol, rtol=rtol, transient=True, final_time=10),
#         subdomains=[my_vol],
#         temperature=300,
#     )
#     H = F.Species("H")
#     my_model.species = [H]

#     my_model.sources = [F.ParticleSource(value=1e20, volume=my_vol, species=H)]
#     my_model.settings.stepsize = F.Stepsize(0.05, milestones=[0.1, 0.2, 0.5, 1])  # s
#     my_model.initialise()

#     my_model.t.value = 0.0
#     my_model.show_progress_bar = False
#     my_model.iterate()
#     # check at t=0
#     assert my_model.solver.atol == atol(t=0.0)
#     assert my_model.solver.rtol == rtol(t=0.0)

#     my_model.t.value = 20
#     my_model.iterate()

#     # check at t=20
#     assert my_model.solver.atol == atol(t=20.0)
#     assert my_model.solver.rtol == rtol(t=20.0)



import numpy as np

from festim.species import Species
from festim.subdomain import VolumeSubdomain


class Profile1DExport:
    """Class to export 1D profiles of a field in a simulation.

    Args:
        field: the species for which the profile is computed.
        subdomain: the volume subdomain to compute the profile on. If None, the profile
            is computed over the entire domain.
        times: if provided, the profile will be exported at these timesteps.
            Otherwise, exports at all timesteps. Defaults to None.

    Attributes:
        field: the species for which the profile is computed.
        subdomain: the volume subdomain to compute the profile on. If None, the profile
            is computed over the entire domain.
        times: if provided, the profile will be exported at these timesteps. Otherwise,
            exports at all timesteps.
        x: the coordinates along which the profile is computed.
        data: the computed profile data.
        t: the list of time values at which the profile is computed.
    """

    x: np.ndarray
    data: list
    field: Species
    subdomain: VolumeSubdomain | None
    times: list[float] | None
    t: list[float]
    _dofs: np.ndarray
    _sort_coords: np.ndarray

    def __init__(
        self,
        field: Species,
        subdomain: VolumeSubdomain = None,
        times: list[float] | None = None,
    ):
        self.field = field
        self.data = []
        self.t = []
        self.x = None
        self.subdomain = subdomain
        self.times = times

        self._dofs = None
        self._sort_coords = None



from mpi4py import MPI

import basix
import dolfinx
import numpy as np
import pytest
from dolfinx import fem

import festim as F

test_mesh_1d = F.Mesh1D(vertices=np.linspace(0, 1, 100))
test_functionspace = fem.functionspace(test_mesh_1d.mesh, ("CG", 1))


@pytest.mark.parametrize(
    "value", ["coucou", 1.0, 1, fem.Constant(test_mesh_1d.mesh, 1.0)]
)
def test_Typeerror_raised_when_wrong_object_given_to_Advection(value):
    "test"

    my_species = F.Species("H")
    my_subdomain = F.VolumeSubdomain(id=1, material="dummy_mat")

    with pytest.raises(
        TypeError,
        match=f"velocity must be a fem.Function, or callable not {type(value)}",
    ):
        F.AdvectionTerm(velocity=value, subdomain=my_subdomain, species=my_species)


@pytest.mark.parametrize(
    "value",
    ["coucou", 1.0, 1, F.SurfaceSubdomain(id=1)],
)
def test_subdomain_setter(value):
    "test"

    my_species = F.Species("H")

    with pytest.raises(
        TypeError,
        match=f"Subdomain must be a festim.Subdomain object, not {type(value)}",
    ):
        F.AdvectionTerm(velocity=None, subdomain=value, species=my_species)


@pytest.mark.parametrize(
    "value",
    ["coucou", 1.0, 1, F.SurfaceSubdomain(id=1)],
)
def test_species_setter_type_error(value):
    "test"

    my_subdomain = F.VolumeSubdomain(id=1, material="dummy_mat")

    with pytest.raises(
        TypeError,
        match=f"elements of species must be of type festim.Species not {type(value)}",
    ):
        F.AdvectionTerm(velocity=None, subdomain=my_subdomain, species=value)


@pytest.mark.parametrize(
    "value",
    [F.Species("H"), [F.Species("D"), F.Species("test")]],
)
def test_species_setter_changes_input_to_list(value):
    "test"

    my_subdomain = F.VolumeSubdomain(id=1, material="dummy_mat")

    my_advection_term = F.AdvectionTerm(
        velocity=None, subdomain=my_subdomain, species=value
    )

    assert isinstance(my_advection_term.species, list)


@pytest.mark.parametrize(
    "value",
    [lambda x: x[0], lambda t: t[0], lambda x, t: x[0] + 2 * t],
)
def test_velocity_accepts_callable_values(value):
    my_subdomain = F.VolumeSubdomain(id=1, material="dummy_mat")

    F.AdvectionTerm(velocity=value, subdomain=my_subdomain, species=F.Species("H"))


def test_subdomain_accepts_None_value():
    F.AdvectionTerm(velocity=None, subdomain=None, species=F.Species("H"))


def test_velocity_field_update():
    """Test a explicit time dependent input value is updated correctly"""

    test_mesh = dolfinx.mesh.create_unit_square(MPI.COMM_WORLD, 10, 10)
    V = fem.functionspace(test_mesh, ("Lagrange", 1))

    v_cg = basix.ufl.element(
        "Lagrange", test_mesh.topology.cell_name(), 2, shape=(test_mesh.geometry.dim,)
    )
    V_velocity = fem.functionspace(test_mesh, v_cg)
    u = fem.Function(V_velocity)

    def velocity_func_alt(x, t):
        values = np.zeros((2, x.shape[1]))  # Initialize with zeros

        scalar_value = 2 * t  # Compute the scalar function
        values[0] = scalar_value  # Assign to first component
        values[1] = scalar_value  # Second component remains zero

        return values

    def example_func(t):
        u.interpolate(lambda x: velocity_func_alt(x, t))
        return u

    my_subdomain = F.VolumeSubdomain(id=1, material="dummy_mat")
    my_species = F.Species("H")
    test_value = F.AdvectionTerm(
        velocity=lambda t: example_func(t), subdomain=my_subdomain, species=my_species
    )
    t = F.as_fenics_constant(value=1.0, mesh=test_mesh)

    test_value.velocity.convert_input_value(function_space=V, t=t)

    t_values = [1, 2, 3, 4, 5]

    for t in t_values:
        test_value.velocity.update(t)
        assert np.isclose(np.max(test_value.velocity.fenics_object.x.array), 2 * t)



import logging
import os

from mpi4py import MPI

import ipyparallel as ipp
import numpy as np
import pytest
from dolfinx import mesh as fenics_mesh
from dolfinx.io import XDMFFile
from dolfinx.mesh import meshtags

import festim as F

mesh_1D = fenics_mesh.create_unit_interval(MPI.COMM_WORLD, 10)
mesh_2D = fenics_mesh.create_unit_square(MPI.COMM_WORLD, 10, 10)
mesh_3D = fenics_mesh.create_unit_cube(MPI.COMM_WORLD, 10, 10, 10)

# 1D meshtags
my_surface_meshtags = meshtags(
    mesh_1D,
    0,
    np.array([0, 10], dtype=np.int32),
    np.array([1, 2], dtype=np.int32),
)

num_cells = mesh_1D.topology.index_map(1).size_local
my_volume_meshtags = meshtags(
    mesh_1D,
    1,
    np.arange(num_cells, dtype=np.int32),
    np.full(num_cells, 1, dtype=np.int32),
)


@pytest.fixture(scope="module")
def cluster():
    cluster = ipp.Cluster(engines="mpi", n=2, log_level=logging.ERROR)
    rc = cluster.start_and_connect_sync()
    yield rc
    cluster.stop_cluster_sync()


@pytest.mark.parametrize("mesh", [mesh_1D, mesh_2D, mesh_3D])
def test_get_fdim(mesh):
    my_mesh = F.Mesh(mesh)

    assert my_mesh.fdim == mesh.topology.dim - 1


def test_fdim_changes_when_mesh_changes():
    my_mesh = F.Mesh(mesh=mesh_1D)

    for mesh in [mesh_1D, mesh_2D, mesh_3D]:
        my_mesh.mesh = mesh
        assert my_mesh.fdim == mesh.topology.dim - 1


@pytest.mark.parametrize("mesh", [mesh_1D, mesh_2D, mesh_3D])
def test_get_vdim(mesh):
    my_mesh = F.Mesh(mesh)

    assert my_mesh.vdim == mesh.topology.dim


def test_vdim_changes_when_mesh_changes():
    my_mesh = F.Mesh(mesh=mesh_1D)

    for mesh in [mesh_1D, mesh_2D, mesh_3D]:
        my_mesh.mesh = mesh
        assert my_mesh.vdim == mesh.topology.dim


@pytest.mark.parametrize("mesh", [mesh_1D, mesh_2D, mesh_3D])
def test_meshtags_from_xdmf(tmp_path, mesh):
    """Test that the facet and volume meshtags are read correctly from the mesh XDMF files"""
    # create mesh functions
    fdim = mesh.topology.dim - 1
    vdim = mesh.topology.dim

    # create facet meshtags
    facet_indices = []
    for i in range(vdim):
        # add the boundary entities at 0 and 1 in each dimension
        facets_zero = fenics_mesh.locate_entities_boundary(
            mesh, fdim, lambda x: np.isclose(x[i], 0)
        )
        facets_one = fenics_mesh.locate_entities_boundary(
            mesh, fdim, lambda x: np.isclose(x[i], 1)
        )

        facet_indices += [facets_zero, facets_one]

    facet_tags = []

    for idx, _ in enumerate(facet_indices):
        # add tags for each boundary
        facet_tag = np.full(len(facet_indices[i]), idx + 1, dtype=np.int32)
        facet_tags.append(facet_tag)

    facet_tags = np.array(facet_tags).flatten()
    facet_indices = np.array(facet_indices).flatten()

    facet_meshtags = fenics_mesh.meshtags(mesh, fdim, facet_indices, facet_tags)

    # create volume meshtags
    num_cells = mesh.topology.index_map(vdim).size_local
    mesh_cell_indices = np.arange(num_cells, dtype=np.int32)
    # tag all volumes with 0
    tags_volumes = np.full(num_cells, 0, dtype=np.int32)
    # create 2 volumes for x<0.5 and x>0.5
    volume_indices_left = fenics_mesh.locate_entities(
        mesh,
        vdim,
        lambda x: x[0] <= 0.5,
    )

    volume_indices_right = fenics_mesh.locate_entities(
        mesh,
        vdim,
        lambda x: x[0] >= 0.5,
    )
    tags_volumes[volume_indices_left] = 2
    tags_volumes[volume_indices_right] = 3

    volume_meshtags = fenics_mesh.meshtags(mesh, vdim, mesh_cell_indices, tags_volumes)

    # write files
    surface_file_path = os.path.join(tmp_path, "facets_file.xdmf")
    surface_file = XDMFFile(MPI.COMM_WORLD, surface_file_path, "w")
    surface_file.write_mesh(mesh)
    surface_file.write_meshtags(facet_meshtags, mesh.geometry)

    volume_file_path = os.path.join(tmp_path, "volumes_file.xdmf")
    volume_file = XDMFFile(MPI.COMM_WORLD, volume_file_path, "w")
    volume_file.write_mesh(mesh)
    volume_file.write_meshtags(volume_meshtags, mesh.geometry)

    # read files
    my_model = F.HydrogenTransportProblem(
        mesh=F.MeshFromXDMF(
            volume_file=volume_file_path,
            facet_file=surface_file_path,
            mesh_name="mesh",
            surface_meshtags_name="mesh_tags",
            volume_meshtags_name="mesh_tags",
        )
    )
    my_model.define_meshtags_and_measures()

    # TEST
    assert volume_meshtags.dim == my_model.volume_meshtags.dim
    assert volume_meshtags.values.all() == my_model.volume_meshtags.values.all()
    assert facet_meshtags.dim == my_model.facet_meshtags.dim
    assert facet_meshtags.values.all() == my_model.facet_meshtags.values.all()


@pytest.mark.parametrize("vertices", [[1, 2, 3, 4], [0, 0.1, 0.2, 0.3, 0.4, 0.5]])
def test_mesh_vertices_from_list(vertices):
    """Check that giving vertices as a list is correctly processed and ends up as a np.ndarray for the mesh"""
    my_mesh = F.Mesh1D(vertices=vertices)

    assert isinstance(my_mesh.vertices, np.ndarray)


def test_error_raised_when_mesh_is_wrong_type():
    """Test that an TypeError is raised when the mesh is not a dolfinx mesh"""

    with pytest.raises(TypeError, match="Mesh must be of type dolfinx.mesh.Mesh"):
        F.Mesh(
            mesh="mesh",
        )


def test_create_1D_mesh_parallel(cluster):
    """Test creating a 1D mesh in parallel using ipyparallel"""

    def create_mesh():
        import numpy as np

        import festim as F

        F.Mesh1D(vertices=np.linspace(0, 1, num=1001))

    query = cluster[:].apply_async(create_mesh)
    query.wait()
    assert query.successful(), query.error


@pytest.mark.parametrize("mesh", [mesh_2D, mesh_3D])
def test_attr_error_with_incompitable_mesh_in_spherical(mesh):
    """Test that an AttributeError is raised when trying to use an incompatible mesh
    with a spherical coordinate system"""

    with pytest.raises(
        AttributeError,
        match="spherical coordinates can be used for one-dimensional domains only",
    ):
        F.Mesh(mesh=mesh, coordinate_system="spherical")


def test_attr_error_with_3D_mesh_in_cylindrical():
    """Test that an AttributeError is raised when trying to use an incompatible mesh
    with a cylindrical coordinate system"""

    with pytest.raises(
        AttributeError,
        match="cylindrical coordinates cannot be used for 3D domains",
    ):
        F.Mesh(mesh=mesh_3D, coordinate_system="cylindrical")


@pytest.mark.parametrize("system", ["cyl", "cart", 1.0, "coucou", mesh_1D])
def test_coordinate_system_setter(system):
    if isinstance(system, str):
        err_msg = "coordinate_system must be one of 'cartesian', 'cylindrical', or 'spherical'"
    else:
        err_msg = "coordinate_system must be of type str or CoordinateSystem"
    with pytest.raises(
        ValueError,
        match=err_msg,
    ):
        F.Mesh(mesh=mesh_3D, coordinate_system=system)



import numpy as np
import ufl
from dolfinx import fem

import festim as F
from festim.subdomain.surface_subdomain import SurfaceSubdomain


class FluxBCBase:
    """
    Flux boundary condition class

    Ensuring the gradient of the solution u at a boundary:

    .. math::

        -A \\nabla u \\cdot \\mathbf{n} = f

    where :math:`A` is some material property (diffusivity for particle flux and thermal
    conductivity for heat flux), :math:`\\mathbf{n}` is the outwards normal vector of
    the boundary, :math:`f` is a function of space and time.


    Args:
        subdomain (festim.SurfaceSubdomain): the surface subdomain where the boundary
            condition is applied
        value (float, fem.Constant, callable): the value of the boundary condition

    Attributes:
        subdomain (festim.SurfaceSubdomain): the surface subdomain where the boundary
            condition is applied
        value (float, fem.Constant, callable): the value of the boundary condition
        value_fenics (fem.Function or fem.Constant): the value of the boundary condition in
            fenics format
        bc_expr (fem.Expression): the expression of the boundary condition that is used to
            update the value_fenics

    """

    def __init__(self, subdomain: SurfaceSubdomain, value):
        self.subdomain = subdomain
        self.value = value

        self.value_fenics = None
        self.bc_expr = None

    @property
    def value_fenics(self):
        return self._value_fenics

    @value_fenics.setter
    def value_fenics(self, value):
        if value is None:
            self._value_fenics = value
            return
        if not isinstance(
            value, fem.Function | fem.Constant | np.ndarray | ufl.core.expr.Expr
        ):
            raise TypeError(
                f"Value must be a dolfinx.fem.Function, dolfinx.fem.Constant, np.ndarray or ufl.core.expr.Expr not {type(value)}"  # noqa: E501
            )
        self._value_fenics = value

    @property
    def time_dependent(self):
        if self.value is None:
            raise TypeError("Value must be given to determine if its time dependent")
        if isinstance(self.value, fem.Constant):
            return False
        if callable(self.value):
            arguments = self.value.__code__.co_varnames
            return "t" in arguments
        else:
            return False

    @property
    def temperature_dependent(self):
        if self.value is None:
            return False
        if isinstance(self.value, fem.Constant):
            return False
        if callable(self.value):
            arguments = self.value.__code__.co_varnames
            return "T" in arguments
        else:
            return False

    def create_value_fenics(self, mesh, temperature, t: fem.Constant):
        """Creates the value of the boundary condition as a fenics object and sets it to
        self.value_fenics.
        If the value is a constant, it is converted to a fenics.Constant.
        If the value is a function of t, it is converted to a fenics.Constant.
        Otherwise, it is converted to a ufl Expression

        Args:
            mesh (dolfinx.mesh.Mesh) : the mesh
            temperature (float): the temperature
            t (dolfinx.fem.Constant): the time
        """
        x = ufl.SpatialCoordinate(mesh)

        if isinstance(self.value, (int, float)):
            self.value_fenics = F.as_fenics_constant(mesh=mesh, value=self.value)

        elif callable(self.value):
            arguments = self.value.__code__.co_varnames

            if "t" in arguments and "x" not in arguments and "T" not in arguments:
                # only t is an argument
                if not isinstance(self.value(t=float(t)), (float, int)):
                    raise ValueError(
                        f"self.value should return a float or an int, not {type(self.value(t=float(t)))} "
                    )
                self.value_fenics = F.as_fenics_constant(
                    mesh=mesh, value=self.value(t=float(t))
                )
            else:
                kwargs = {}
                if "t" in arguments:
                    kwargs["t"] = t
                if "x" in arguments:
                    kwargs["x"] = x
                if "T" in arguments:
                    kwargs["T"] = temperature

                self.value_fenics = self.value(**kwargs)

    def update(self, t):
        """Updates the flux bc value

        Args:
            t (float): the time
        """
        if callable(self.value):
            arguments = self.value.__code__.co_varnames
            if isinstance(self.value_fenics, fem.Constant) and "t" in arguments:
                self.value_fenics.value = self.value(t=t)


class ParticleFluxBC(FluxBCBase):
    """
    Particle flux boundary condition class
    Ensuring the gradient of the solution c at a boundary:
    -D * grad(c) * n = f
    where D is the material diffusivity, n is the outwards normal vector of the boundary, f is a function of space and time.

    Args:
        subdomain (festim.SurfaceSubdomain): the surface subdomain where the particle flux
            is applied
        value (float, fem.Constant, callable): the value of the particle flux
        species (festim.Species): the species to which the flux is applied
        species_dependent_value (dict): a dictionary containing the species that the value. Example: {"name": species}
            where "name" is the variable name in the callable value and species is a festim.Species object.

    Attributes:
        subdomain (festim.SurfaceSubdomain): the surface subdomain where the particle flux
            is applied
        value (float or fem.Constant): the value of the particle flux
        species (festim.Species): the species to which the flux is applied
        value_fenics (fem.Function or fem.Constant): the value of the particle flux in
            fenics format
        bc_expr (fem.Expression): the expression of the particle flux that is used to
            update the value_fenics
        species_dependent_value (dict): a dictionary containing the species that the value. Example: {"name": species}
            where "name" is the variable name in the callable value and species is a festim.Species object.


    Examples:

        .. testsetup:: ParticleFluxBC

            from festim import SurfaceSubdomain, ParticleFluxBC, Species
            my_subdomain = SurfaceSubdomain(id=1)
            species1 = Species(name="1")

        .. testcode:: ParticleFluxBC

            ParticleFluxBC(subdomain=my_subdomain, value=1, species="H")
            ParticleFluxBC(subdomain=my_subdomain, value=lambda x: 1 + x[0], species="H")
            ParticleFluxBC(subdomain=my_subdomain, value=lambda t: 1 + t, species="H")
            ParticleFluxBC(subdomain=my_subdomain, value=lambda T: 1 + T, species="H")
            ParticleFluxBC(subdomain=my_subdomain, value=lambda x, t: 1 + x[0] + t, species="H")
            ParticleFluxBC(subdomain=my_subdomain, value=lambda c1: 2 * c1**2, species="H", species_dependent_value={"c1": species1})
    """

    def __init__(self, subdomain, value, species, species_dependent_value={}):
        super().__init__(subdomain=subdomain, value=value)
        self.species = species
        self.species_dependent_value = species_dependent_value

    def create_value_fenics(self, mesh, temperature, t: fem.Constant):
        """Creates the value of the boundary condition as a fenics object and sets it to
        self.value_fenics.
        If the value is a constant, it is converted to a fenics.Constant.
        If the value is a function of t, it is converted to a fenics.Constant.
        Otherwise, it is converted to a ufl Expression

        Args:
            mesh (dolfinx.mesh.Mesh) : the mesh
            temperature (float): the temperature
            t (dolfinx.fem.Constant): the time
        """
        x = ufl.SpatialCoordinate(mesh)

        if isinstance(self.value, (int, float)):
            self.value_fenics = F.as_fenics_constant(mesh=mesh, value=self.value)

        elif callable(self.value):
            arguments = self.value.__code__.co_varnames

            if "t" in arguments and "x" not in arguments and "T" not in arguments:
                # only t is an argument
                if not isinstance(self.value(t=float(t)), (float, int)):
                    raise ValueError(
                        f"self.value should return a float or an int, not {type(self.value(t=float(t)))} "
                    )
                self.value_fenics = F.as_fenics_constant(
                    mesh=mesh, value=self.value(t=float(t))
                )
            else:
                kwargs = {}
                if "t" in arguments:
                    kwargs["t"] = t
                if "x" in arguments:
                    kwargs["x"] = x
                if "T" in arguments:
                    kwargs["T"] = temperature

                for name, species in self.species_dependent_value.items():
                    kwargs[name] = species.concentration

                self.value_fenics = self.value(**kwargs)


class HeatFluxBC(FluxBCBase):
    """
    Heat flux boundary condition class
    Ensuring the gradient of the solution T at a boundary:
    -lambda * grad(T) * n = f
    where lambda is the thermal conductivity , n is the outwards normal vector of the boundary, f is a function of space and time.

    Args:
        subdomain (festim.SurfaceSubdomain): the surface subdomain where the heat flux
            is applied
        value (float, callable, fem.Constant): the value of the heat flux

    Attributes:
        subdomain (festim.SurfaceSubdomain): the surface subdomain where the heat flux
            is applied
        value (float, callable, fem.Constant): the value of the heat flux
        value_fenics (fem.Function or fem.Constant): the value of the heat flux in
            fenics format
        bc_expr (fem.Expression): the expression of the heat flux that is used to
            update the value_fenics


    Examples:

        .. testsetup:: HeatFluxBC

                from festim import SurfaceSubdomain, HeatFluxBC, Species
                my_subdomain = SurfaceSubdomain(id=1)
                species1 = Species(name="1")

        .. testcode:: HeatFluxBC

            HeatFluxBC(subdomain=my_subdomain, value=1)
            HeatFluxBC(subdomain=my_subdomain, value=lambda x: 1 + x[0])
            HeatFluxBC(subdomain=my_subdomain, value=lambda t: 1 + t)
            HeatFluxBC(subdomain=my_subdomain, value=lambda x, t: 1 + x[0] + t)
    """

    def __init__(self, subdomain, value):
        super().__init__(subdomain=subdomain, value=value)



import numpy as np

import festim as F


def test_sim_reaction_not_in_every_volume():
    """Tests that a steady simulation can be run if a reaction is not defined
    in every volume"""
    my_model = F.HydrogenTransportProblem()
    my_model.mesh = F.Mesh1D(np.linspace(0, 2e-04, num=1001))

    my_mat = F.Material(D_0=4.1e-07, E_D=0.37, name="my_mat")
    vol_1 = F.VolumeSubdomain1D(id=1, borders=[0, 1e-04], material=my_mat)
    vol_2 = F.VolumeSubdomain1D(id=2, borders=[1e-04, 2e-04], material=my_mat)
    left = F.SurfaceSubdomain1D(id=1, x=0)
    my_model.subdomains = [vol_1, vol_2, left]

    cm, ct = F.Species("cm"), F.Species("ct", mobile=False)
    my_model.species = [cm, ct]
    trap = F.ImplicitSpecies(n=8.19e25, others=[ct])
    my_model.reactions = [
        F.Reaction(
            reactant=[cm, trap],
            product=ct,
            k_0=8.9e-17,
            E_k=0.39,
            p_0=1e13,
            E_p=0.87,
            volume=vol_1,
        ),
    ]

    my_model.temperature = 600

    my_model.settings = F.Settings(atol=1e10, rtol=1e-10, transient=False)

    my_model.initialise()
    my_model.run()



from petsc4py import PETSc

import dolfinx
import numpy as np
from dolfinx.fem import Constant
from packaging.version import Version
from ufl import exp

import festim as F


def relative_error_computed_to_analytical(
    D, permeability, computed_flux, L, times, P_up
):
    n_array = np.arange(1, 10000)[:, np.newaxis]
    summation = np.sum(
        (-1) ** n_array * np.exp(-((np.pi * n_array) ** 2) * float(D) / L**2 * times),
        axis=0,
    )
    analytical_flux = P_up**0.5 * permeability / L * (2 * summation + 1)

    # post processing
    analytical_flux = np.abs(analytical_flux)
    indices = np.where(analytical_flux > 0.1 * np.max(analytical_flux))
    analytical_flux = analytical_flux[indices]
    computed_flux = computed_flux[indices]

    # evaulate relative error compared to analytical solution
    relative_error = np.abs((computed_flux - analytical_flux) / analytical_flux)
    error = relative_error.mean()

    return error


def test_multispecies_permeation_problem():
    """Test running a problem with 2 mobile species permeating through a 1D
    domain, with different diffusion coefficients, checks that the computed
    permeation flux matches the analytical solution"""

    # festim model
    L = 3e-04
    my_mesh = F.Mesh1D(np.linspace(0, L, num=1001))
    my_model = F.HydrogenTransportProblem()
    my_model.mesh = my_mesh

    my_mat = F.Material(
        D_0={"spe_1": 1.9e-7, "spe_2": 3.8e-7},
        E_D={"spe_1": 0.2, "spe_2": 0.2},
        name="my_mat",
    )
    my_subdomain = F.VolumeSubdomain1D(id=1, borders=[0, L], material=my_mat)
    left_surface = F.SurfaceSubdomain1D(id=1, x=0)
    right_surface = F.SurfaceSubdomain1D(id=2, x=L)
    my_model.subdomains = [
        my_subdomain,
        left_surface,
        right_surface,
    ]

    spe_1 = F.Species("spe_1")
    spe_2 = F.Species("spe_2")
    my_model.species = [spe_1, spe_2]

    temperature = Constant(my_mesh.mesh, 500.0)
    my_model.temperature = temperature
    pressure = 100

    my_model.boundary_conditions = [
        F.DirichletBC(subdomain=right_surface, value=0, species="spe_1"),
        F.DirichletBC(subdomain=right_surface, value=0, species="spe_2"),
        F.SievertsBC(
            subdomain=left_surface,
            S_0=4.02e21,
            E_S=1.04,
            pressure=pressure,
            species="spe_1",
        ),
        F.SievertsBC(
            subdomain=left_surface,
            S_0=5.0e21,
            E_S=1.2,
            pressure=pressure,
            species="spe_2",
        ),
    ]
    my_model.settings = F.Settings(
        atol=1e10,
        rtol=1e-10,
        max_iterations=30,
        final_time=10,
    )
    my_model.settings.stepsize = F.Stepsize(initial_value=1 / 20)

    outgassing_flux_spe_1 = F.SurfaceFlux(
        field=spe_1,
        surface=right_surface,
    )
    outgassing_flux_spe_2 = F.SurfaceFlux(
        field=spe_2,
        surface=right_surface,
    )
    total_species_1 = F.TotalVolume(
        field=spe_1,
        volume=my_subdomain,
    )
    total_species_2 = F.TotalVolume(
        field=spe_2,
        volume=my_subdomain,
    )
    my_model.exports = [
        outgassing_flux_spe_1,
        outgassing_flux_spe_2,
        total_species_1,
        total_species_2,
    ]
    my_model.initialise()

    if Version(dolfinx.__version__) == Version("0.9.0"):
        my_model.solver.convergence_criterion = "incremental"
        ksp = my_model.solver.krylov_solver
        opts = PETSc.Options()
        option_prefix = ksp.getOptionsPrefix()
        opts[f"{option_prefix}ksp_type"] = "cg"
        opts[f"{option_prefix}pc_type"] = "gamg"
        opts[f"{option_prefix}pc_factor_mat_solver_type"] = "mumps"
        ksp.setFromOptions()
    elif Version(dolfinx.__version__) > Version("0.9.0"):
        snes = my_model.solver.solver
        opts = PETSc.Options()
        option_prefix = snes.getOptionsPrefix()
        opts[f"{option_prefix}snes_atol"] = 0
        opts[f"{option_prefix}snes_rtol"] = 0
        opts[f"{option_prefix}snes_stol"] = 1e-8
        opts[f"{option_prefix}ksp_type"] = "cg"
        opts[f"{option_prefix}pc_type"] = "gamg"
        opts[f"{option_prefix}pc_factor_mat_solver_type"] = "mumps"
        snes.setFromOptions()
    my_model.run()

    times = outgassing_flux_spe_1.t
    flux_values_spe_1 = outgassing_flux_spe_1.data
    flux_values_spe_2 = outgassing_flux_spe_2.data

    # ---------------------- analytical solutions -----------------------------

    # common values
    times = np.array(times)
    P_up = float(my_model.boundary_conditions[-1].pressure)

    # ##### compute analyical solution for species 1 ##### #
    D_spe_1 = my_mat.get_diffusion_coefficient(my_mesh.mesh, temperature, spe_1)
    S_0_spe_1 = float(my_model.boundary_conditions[-2].S_0)
    E_S_spe_1 = float(my_model.boundary_conditions[-2].E_S)
    S_spe_1 = S_0_spe_1 * exp(-E_S_spe_1 / F.k_B / float(temperature))
    permeability_spe_1 = float(D_spe_1) * S_spe_1
    flux_values_spe_1 = np.array(np.abs(flux_values_spe_1))

    error_spe_1 = relative_error_computed_to_analytical(
        D_spe_1, permeability_spe_1, flux_values_spe_1, L, times, P_up
    )

    # ##### compute analyical solution for species 2 ##### #
    D_spe_2 = my_mat.get_diffusion_coefficient(my_mesh.mesh, temperature, spe_2)
    S_0_spe_2 = float(my_model.boundary_conditions[-1].S_0)
    E_S_spe_2 = float(my_model.boundary_conditions[-1].E_S)
    S_spe_2 = S_0_spe_2 * exp(-E_S_spe_2 / F.k_B / float(temperature))
    permeability_spe_2 = float(D_spe_2) * S_spe_2
    flux_values_spe_2 = np.array(np.abs(flux_values_spe_2))

    error_spe_2 = relative_error_computed_to_analytical(
        D_spe_2, permeability_spe_2, flux_values_spe_2, L, times, P_up
    )

    for err in [error_spe_1, error_spe_2]:
        assert err < 0.01



import numpy as np
import ufl
from dolfinx import fem

import festim as F


def test_total_surface_compute_1D():
    """Test that the total surface export computes the correct value"""

    # BUILD
    L = 6
    D = 1.5
    my_mesh = F.Mesh1D(np.linspace(0, L, 10000))
    dummy_surface = F.SurfaceSubdomain1D(id=1, x=6)
    dummy_volume = F.VolumeSubdomain1D(
        id=1, borders=[0, L], material=F.Material(D_0=1, E_D=1, name="dummy")
    )
    facet_meshtags, temp = my_mesh.define_meshtags(
        surface_subdomains=[dummy_surface], volume_subdomains=[dummy_volume]
    )
    ds = ufl.Measure("ds", domain=my_mesh.mesh, subdomain_data=facet_meshtags)

    # give function to species
    V = fem.functionspace(my_mesh.mesh, ("Lagrange", 1))
    c = fem.Function(V)
    c.interpolate(lambda x: x[0] * 0.5 + 1)

    my_species = F.Species("H")
    my_species.solution = c

    my_export = F.TotalSurface(field=my_species, surface=dummy_surface)
    my_export.D = D

    # RUN
    my_export.compute(ds)

    # TEST
    expected_value = 4.0
    computed_value = my_export.value

    assert np.isclose(computed_value, expected_value, rtol=1e-2)



from mpi4py import MPI

import numpy as np
import ufl
from dolfinx import fem
from dolfinx.mesh import create_unit_cube, create_unit_square, locate_entities

import festim as F

from .tools import error_L2

test_mesh_1d = F.Mesh1D(np.linspace(0, 1, 10000))
test_mesh_2d = create_unit_square(MPI.COMM_WORLD, 50, 50)
test_mesh_3d = create_unit_cube(MPI.COMM_WORLD, 20, 20, 20)
x_1d = ufl.SpatialCoordinate(test_mesh_1d.mesh)
x_2d = ufl.SpatialCoordinate(test_mesh_2d)
x_3d = ufl.SpatialCoordinate(test_mesh_3d)


def test_1_mobile_MMS_steady_state():
    """
    MMS test with one mobile species at steady state
    """

    def u_exact(mod):
        return lambda x: 1 + mod.sin(2 * mod.pi * x[0])

    H_analytical_ufl = u_exact(ufl)
    H_analytical_np = u_exact(np)

    V = fem.functionspace(test_mesh_1d.mesh, ("Lagrange", 1))
    T = fem.Function(V)

    D_0 = 1
    E_D = 0.1
    T_expr = lambda x: 500 + 100 * x[0]
    T.interpolate(T_expr)
    D = D_0 * ufl.exp(-E_D / (F.k_B * T))

    my_model = F.HydrogenTransportProblem()
    my_model.mesh = test_mesh_1d
    my_mat = F.Material(name="mat", D_0=D_0, E_D=E_D)
    vol = F.VolumeSubdomain1D(id=1, borders=[0, 1], material=my_mat)
    left = F.SurfaceSubdomain1D(id=1, x=0)
    right = F.SurfaceSubdomain1D(id=2, x=1)

    my_model.subdomains = [vol, left, right]

    H = F.Species("H")
    my_model.species = [H]

    my_model.temperature = T_expr

    my_model.boundary_conditions = [
        F.DirichletBC(subdomain=left, value=H_analytical_ufl, species=H),
        F.DirichletBC(subdomain=right, value=H_analytical_ufl, species=H),
    ]

    f = -ufl.div(D * ufl.grad(H_analytical_ufl(x_1d)))
    my_model.sources = [F.ParticleSource(value=f, volume=vol, species=H)]

    my_model.settings = F.Settings(atol=1e-10, rtol=1e-10, transient=False)

    my_model.initialise()
    my_model.run()

    H_computed = H.post_processing_solution

    L2_error = error_L2(H_computed, H_analytical_np)

    assert L2_error < 1e-7


def test_1_mobile_MMS_transient():
    """
    MMS test with 1 mobile species in 0.1s transient, the value at the last time step is
    compared to an analytical solution
    """

    final_time = 0.1

    def u_exact(mod):
        return lambda x, t: 1 + mod.sin(2 * mod.pi * x[0]) + 2 * t**2

    def u_exact_alt(mod):
        return lambda x: u_exact(mod)(x, final_time)

    H_analytical_ufl = u_exact(ufl)
    H_analytical_np = u_exact_alt(np)

    V = fem.functionspace(test_mesh_1d.mesh, ("Lagrange", 1))
    T = fem.Function(V)

    D_0 = 1
    E_D = 0.1
    T_expr = lambda x: 600 + 50 * x[0]
    T.interpolate(T_expr)
    D = D_0 * ufl.exp(-E_D / (F.k_B * T))

    # FESTIM model

    my_model = F.HydrogenTransportProblem()
    my_model.mesh = test_mesh_1d
    my_mat = F.Material(name="mat", D_0=D_0, E_D=E_D)
    vol = F.VolumeSubdomain1D(id=1, borders=[0, 1], material=my_mat)
    left = F.SurfaceSubdomain1D(id=1, x=0)
    right = F.SurfaceSubdomain1D(id=2, x=1)

    my_model.subdomains = [vol, left, right]

    H = F.Species("H")
    my_model.species = [H]

    my_model.temperature = T_expr

    my_model.boundary_conditions = [
        F.DirichletBC(subdomain=left, value=H_analytical_ufl, species=H),
        F.DirichletBC(subdomain=right, value=H_analytical_ufl, species=H),
    ]

    init_value = lambda x: 1 + ufl.sin(2 * ufl.pi * x[0])
    my_model.initial_conditions = [
        F.InitialConcentration(value=init_value, species=H, volume=vol)
    ]

    f = lambda x, t: 4 * t - ufl.div(D * ufl.grad(H_analytical_ufl(x, t)))
    my_model.sources = [F.ParticleSource(value=f, volume=vol, species=H)]

    my_model.settings = F.Settings(atol=1e-10, rtol=1e-10, final_time=final_time)
    my_model.settings.stepsize = final_time / 50

    my_model.initialise()
    my_model.run()

    H_computed = H.post_processing_solution

    L2_error = error_L2(H_computed, H_analytical_np)

    assert L2_error < 5e-4


def test_1_mobile_MMS_2D():
    """Tests that a steady simulation can be run in a 2D domain with
    1 mobile species"""

    def u_exact(mod):
        return lambda x: 1 + mod.sin(2 * mod.pi * x[0]) + mod.cos(2 * mod.pi * x[1])

    H_analytical_ufl = u_exact(ufl)
    H_analytical_np = u_exact(np)

    V = fem.functionspace(test_mesh_2d, ("Lagrange", 1))
    T = fem.Function(V)

    D_0 = 1
    E_D = 0.1
    T_expr = lambda x: 500 + 100 * x[0]
    T.interpolate(T_expr)
    D = D_0 * ufl.exp(-E_D / (F.k_B * T))

    my_model = F.HydrogenTransportProblem()
    my_model.mesh = F.Mesh(mesh=test_mesh_2d)

    class LefSurface(F.SurfaceSubdomain):
        def locate_boundary_facet_indices(self, mesh):
            fdim = mesh.topology.dim - 1
            indices = locate_entities(mesh, fdim, lambda x: np.isclose(x[0], 0))
            return indices

    class RightSurface(F.SurfaceSubdomain):
        def locate_boundary_facet_indices(self, mesh):
            fdim = mesh.topology.dim - 1
            indices = locate_entities(mesh, fdim, lambda x: np.isclose(x[0], 1))
            return indices

    my_mat = F.Material(name="mat", D_0=D_0, E_D=E_D)
    vol = F.VolumeSubdomain(id=1, material=my_mat)
    left = LefSurface(id=1)
    right = RightSurface(id=2)
    my_model.subdomains = [vol, left, right]

    H = F.Species("H")
    my_model.species = [H]

    my_model.temperature = T_expr

    my_model.boundary_conditions = [
        F.DirichletBC(subdomain=left, value=H_analytical_ufl, species=H),
        F.DirichletBC(subdomain=right, value=H_analytical_ufl, species=H),
    ]

    f = -ufl.div(D * ufl.grad(H_analytical_ufl(x_2d)))
    my_model.sources = [F.ParticleSource(value=f, volume=vol, species=H)]

    my_model.settings = F.Settings(atol=1e-10, rtol=1e-10, transient=False)

    my_model.initialise()
    my_model.run()

    H_computed = H.post_processing_solution

    L2_error = error_L2(H_computed, H_analytical_np)

    assert L2_error < 2e-3


def test_1_mobile_MMS_3D():
    """Tests that a steady simulation can be run in a 3D domain with
    1 mobile species"""

    def u_exact(mod):
        return lambda x: 1 + mod.sin(2 * mod.pi * x[0]) + mod.cos(2 * mod.pi * x[1])

    H_analytical_ufl = u_exact(ufl)
    H_analytical_np = u_exact(np)

    V = fem.functionspace(test_mesh_3d, ("Lagrange", 1))
    T = fem.Function(V)

    D_0 = 1
    E_D = 0.1
    T_expr = lambda x: 500 + 100 * x[0]
    T.interpolate(T_expr)
    D = D_0 * ufl.exp(-E_D / (F.k_B * T))

    my_model = F.HydrogenTransportProblem()
    my_model.mesh = F.Mesh(mesh=test_mesh_3d)

    class LefSurface(F.SurfaceSubdomain):
        def locate_boundary_facet_indices(self, mesh):
            fdim = mesh.topology.dim - 1
            indices = locate_entities(mesh, fdim, lambda x: np.isclose(x[0], 0))
            return indices

    class RightSurface(F.SurfaceSubdomain):
        def locate_boundary_facet_indices(self, mesh):
            fdim = mesh.topology.dim - 1
            indices = locate_entities(mesh, fdim, lambda x: np.isclose(x[0], 1))
            return indices

    my_mat = F.Material(name="mat", D_0=D_0, E_D=E_D)
    vol = F.VolumeSubdomain(id=1, material=my_mat)
    left = LefSurface(id=1)
    right = RightSurface(id=2)

    my_model.subdomains = [vol, left, right]

    H = F.Species("H")
    my_model.species = [H]

    my_model.temperature = T_expr

    my_model.boundary_conditions = [
        F.DirichletBC(subdomain=left, value=H_analytical_ufl, species=H),
        F.DirichletBC(subdomain=right, value=H_analytical_ufl, species=H),
    ]

    f = -ufl.div(D * ufl.grad(H_analytical_ufl(x_3d)))
    my_model.sources = [F.ParticleSource(value=f, volume=vol, species=H)]

    my_model.settings = F.Settings(atol=1e-10, rtol=1e-10, transient=False)

    my_model.initialise()
    my_model.run()

    H_computed = H.post_processing_solution

    L2_error = error_L2(H_computed, H_analytical_np)

    assert L2_error < 1e-2



from collections.abc import Callable

import numpy as np
import numpy.typing as npt
import ufl
import ufl.core
import ufl.core.expr
from dolfinx import fem
from dolfinx import mesh as _mesh

from festim import helpers
from festim import subdomain as _subdomain
from festim.species import Species


class DirichletBCBase:
    """
    Dirichlet boundary condition class
    u = value

    Args:
        subdomain: The surface subdomain where the boundary condition is applied
        value: The value of the boundary condition

    Attributes:
        subdomain: The surface subdomain where the boundary condition is applied
        value: The value of the boundary condition
        value_fenics: The value of the boundary condition in fenics format
        bc_expr: The expression of the boundary condition that is used to
            update the `value_fenics`

    """

    subdomain: _subdomain.SurfaceSubdomain
    value: (
        np.ndarray
        | fem.Constant
        | int
        | float
        | Callable[[np.ndarray], np.ndarray]
        | Callable[[np.ndarray, float], np.ndarray]
        | Callable[[float], float]
    )
    value_fenics: None | fem.Function | fem.Constant | np.ndarray | float
    bc_expr: fem.Expression

    def __init__(
        self,
        subdomain: _subdomain.SurfaceSubdomain,
        value: np.ndarray | fem.Constant | int | float | Callable,
    ):
        self.subdomain = subdomain
        self.value = value

        self.value_fenics = None
        self.bc_expr = None

    @property
    def value_fenics(self):
        return self._value_fenics

    @value_fenics.setter
    def value_fenics(self, value: None | fem.Function | fem.Constant | np.ndarray):
        if value is None:
            self._value_fenics = value
            return
        if not isinstance(value, (fem.Function, fem.Constant, np.ndarray)):
            # FIXME: Should we allow sending in a callable here?
            raise TypeError(
                "Value must be a dolfinx.fem.Function, dolfinx.fem.Constant, or a np.ndarray not"
                + f"{type(value)}"
            )
        self._value_fenics = value

    @property
    def time_dependent(self) -> bool:
        """Returns true if the value of the boundary condition is time dependent"""
        if self.value is None:
            return False
        if isinstance(self.value, fem.Constant):
            return False
        if callable(self.value):
            arguments = self.value.__code__.co_varnames
            return "t" in arguments
        else:
            return False

    def define_surface_subdomain_dofs(
        self,
        facet_meshtags: _mesh.MeshTags,
        function_space: fem.FunctionSpace | tuple[fem.FunctionSpace, fem.FunctionSpace],
    ) -> npt.NDArray[np.int32] | tuple[npt.NDArray[np.int32], npt.NDArray[np.int32]]:
        """Defines the facets and the degrees of freedom of the boundary condition.

        Given the input meshtags, find all facets matching the boundary condition subdomain ID,
        and locate all DOFs associated with the input function space(s).

        Note:
            For sub-spaces, a tuple of sub-spaces are expected as input, and a tuple of arrays
            associated to each of the function spaces are returned.

        Args:
            facet_meshtags: MeshTags describing some facets in the domain
            mesh:
            function_space: The function space or a tuple of function spaces: (sub, collapsed)
        """
        mesh = (
            function_space[0].mesh
            if isinstance(function_space, tuple)
            else function_space.mesh
        )
        if facet_meshtags.topology != mesh.topology._cpp_object:
            raise ValueError(
                "Mesh of function-space is not the same as the one used for the meshtags"
            )
        if mesh.topology.dim - 1 != facet_meshtags.dim:
            raise ValueError(
                f"Meshtags of dimension {facet_meshtags.dim}, expected {mesh.topology.dim - 1}"
            )
        bc_dofs = fem.locate_dofs_topological(
            function_space, facet_meshtags.dim, facet_meshtags.find(self.subdomain.id)
        )

        return bc_dofs

    def update(self, t: float):
        """Updates the boundary condition value

        Args:
            t: the time
        """
        if callable(self.value):
            arguments = self.value.__code__.co_varnames
            if isinstance(self.value_fenics, fem.Constant) and "t" in arguments:
                self.value_fenics.value = self.value(t=t)
            else:
                self.value_fenics.interpolate(self.bc_expr)
        elif self.bc_expr is not None:
            self.value_fenics.interpolate(self.bc_expr)


class FixedConcentrationBC(DirichletBCBase):
    """
    Args:
        subdomain (festim.Subdomain): the surface subdomain where the boundary
            condition is applied
        value: The value of the boundary condition. It can be a function of space and/or time
        species: The name of the species

    Attributes:
        temperature_dependent (bool): True if the value of the bc is temperature dependent

    Examples:

        .. testsetup:: FixedConcentrationBC

            from festim import FixedConcentrationBC, SurfaceSubdomain
            my_subdomain = SurfaceSubdomain(id=1)

        .. testcode:: FixedConcentrationBC

            FixedConcentrationBC(subdomain=my_subdomain, value=1, species="H")
            FixedConcentrationBC(subdomain=my_subdomain,
                                 value=lambda x: 1 + x[0], species="H")
            FixedConcentrationBC(subdomain=my_subdomain,
                                 value=lambda t: 1 + t, species="H")
            FixedConcentrationBC(subdomain=my_subdomain,
                                 value=lambda T: 1 + T, species="H")
            FixedConcentrationBC(subdomain=my_subdomain,
                                 value=lambda x, t: 1 + x[0] + t, species="H")

    """

    species: Species

    def __init__(
        self,
        subdomain: _subdomain.SurfaceSubdomain,
        value: np.ndarray | fem.Constant | int | float | Callable,
        species: Species,
    ):
        self.species = species
        super().__init__(subdomain, value)

    @property
    def temperature_dependent(self):
        if self.value is None:
            return False
        if isinstance(self.value, fem.Constant):
            return False
        if callable(self.value):
            arguments = self.value.__code__.co_varnames
            return "T" in arguments
        else:
            return False

    def create_value(
        self,
        function_space: fem.FunctionSpace,
        temperature: float | fem.Constant,
        t: float | fem.Constant,
        K_S: fem.Function = None,
    ):
        """Creates the value of the boundary condition as a fenics object and sets it to
        self.value_fenics.
        If the value is a constant, it is converted to a `dolfinx.fem.Constant`.
        If the value is a function of t, it is converted to  `dolfinx.fem.Constant`.
        Otherwise, it is converted to a `dolfinx.fem.Function`.Function and the
        expression of the function is stored in `bc_expr`.

        Args:
            function_space: the function space
            temperature: The temperature
            t: the time
            K_S: The solubility of the species. If provided, the value of the boundary condition
                is divided by K_S (change of variable method).
        """
        mesh = function_space.mesh
        x = ufl.SpatialCoordinate(mesh)

        if isinstance(self.value, (int, float)):
            self.value_fenics = helpers.as_fenics_constant(mesh=mesh, value=self.value)

        elif callable(self.value):
            arguments = self.value.__code__.co_varnames

            if "t" in arguments and "x" not in arguments and "T" not in arguments:
                # only t is an argument
                if not isinstance(self.value(t=float(t)), (float, int)):
                    raise ValueError(
                        "self.value should return a float or an int, not "
                        + f"{type(self.value(t=float(t)))} "
                    )
                self.value_fenics = helpers.as_fenics_constant(
                    mesh=mesh, value=self.value(t=float(t))
                )
            else:
                kwargs = {}
                if "t" in arguments:
                    kwargs["t"] = t
                if "x" in arguments:
                    kwargs["x"] = x
                if "T" in arguments:
                    kwargs["T"] = temperature

                try:
                    self.value_fenics = fem.Function(function_space)

                    # store the expression of the boundary condition
                    # to update the value_fenics later
                    self.bc_expr = fem.Expression(
                        self.value(**kwargs),
                        helpers.get_interpolation_points(function_space.element),
                    )
                    self.value_fenics.interpolate(self.bc_expr)
                except RuntimeError:
                    # if this fails, it is probably because the temperature is a Function
                    # from the parent mesh and this is used in a mixed domain problem.
                    # In this case, we need to interpolate the temperature on the submesh

                    submesh = mesh
                    mesh_cell_map = submesh.topology.index_map(submesh.topology.dim)
                    num_cells_on_proc = (
                        mesh_cell_map.size_local + mesh_cell_map.num_ghosts
                    )
                    cells = np.arange(num_cells_on_proc, dtype=np.int32)
                    parent_functionspace = temperature.function_space
                    interpolation_data = fem.create_interpolation_data(
                        function_space, parent_functionspace, cells
                    )

                    temperature_sub = fem.Function(function_space)
                    temperature_sub.interpolate_nonmatching(
                        temperature,
                        cells=cells,
                        interpolation_data=interpolation_data,
                    )

                    # override the kwargs with the temperature_sub
                    kwargs["T"] = temperature_sub

                    self.bc_expr = fem.Expression(
                        self.value(**kwargs),
                        helpers.get_interpolation_points(function_space.element),
                    )
                    self.value_fenics.interpolate(self.bc_expr)

        # if K_S is provided, divide the value by K_S (change of variable method)
        if K_S is not None:
            if isinstance(self.value, (int, float)):
                val_as_cst = helpers.as_fenics_constant(mesh=mesh, value=self.value)
                self.bc_expr = fem.Expression(
                    val_as_cst / K_S,
                    helpers.get_interpolation_points(function_space.element),
                )
                self.value_fenics = fem.Function(function_space)
                self.value_fenics.interpolate(self.bc_expr)

            elif callable(self.value):
                arguments = self.value.__code__.co_varnames

                if "t" in arguments and "x" not in arguments and "T" not in arguments:
                    # only t is an argument

                    # check that value returns a ufl expression
                    if not isinstance(self.value(t=t), (ufl.core.expr.Expr)):
                        raise ValueError(
                            "self.value should return a ufl expression"
                            + f"{type(self.value(t=t))} "
                        )

                    self.bc_expr = fem.Expression(
                        self.value(t=t) / K_S,
                        helpers.get_interpolation_points(function_space.element),
                    )
                    self.value_fenics = fem.Function(function_space)
                    self.value_fenics.interpolate(self.bc_expr)
                else:
                    self.value_fenics = fem.Function(function_space)
                    kwargs = {}
                    if "t" in arguments:
                        kwargs["t"] = t
                    if "x" in arguments:
                        kwargs["x"] = x
                    if "T" in arguments:
                        kwargs["T"] = temperature

                    # store the expression of the boundary condition
                    # to update the value_fenics later
                    self.bc_expr = fem.Expression(
                        self.value(**kwargs) / K_S,
                        helpers.get_interpolation_points(function_space.element),
                    )
                    self.value_fenics.interpolate(self.bc_expr)


# alias for FixedConcentrationBC
DirichletBC = FixedConcentrationBC


class FixedTemperatureBC(DirichletBCBase):
    def create_value(self, function_space: fem.FunctionSpace, t: fem.Constant):
        """Creates the value of the boundary condition as a fenics object and sets it to
        self.value_fenics.
        If the value is a constant, it is converted to a `dolfinx.fem.Constant`.
        If the value is a function of t, it is converted to a `dolfinx.fem.Constant`.
        Otherwise, it is converted to a` dolfinx.fem.Function` and the
        expression of the function is stored in `bc_expr`.

        Args:
            function_space: the function space
            t: the time
        """
        mesh = function_space.mesh
        x = ufl.SpatialCoordinate(mesh)

        if isinstance(self.value, (int, float)):
            self.value_fenics = helpers.as_fenics_constant(mesh=mesh, value=self.value)

        elif callable(self.value):
            arguments = self.value.__code__.co_varnames

            if "t" in arguments and "x" not in arguments:
                # only t is an argument
                if not isinstance(self.value(t=float(t)), (float, int)):
                    raise ValueError(
                        "self.value should return a float or an int, not "
                        + f"{type(self.value(t=float(t)))} "
                    )
                self.value_fenics = helpers.as_fenics_constant(
                    mesh=mesh, value=self.value(t=float(t))
                )
            else:
                self.value_fenics = fem.Function(function_space)
                kwargs = {}
                if "t" in arguments:
                    kwargs["t"] = t
                if "x" in arguments:
                    kwargs["x"] = x

                # store the expression of the boundary condition
                # to update the value_fenics later
                self.bc_expr = fem.Expression(
                    self.value(**kwargs),
                    helpers.get_interpolation_points(function_space.element),
                )
                self.value_fenics.interpolate(self.bc_expr)



from .interface import Interface
from .surface_subdomain import SurfaceSubdomain, SurfaceSubdomain1D
from .volume_subdomain import VolumeSubdomain, VolumeSubdomain1D

__all__ = [
    "Interface",
    "Subdomain",
    "SurfaceSubdomain",
    "SurfaceSubdomain1D",
    "VolumeSubdomain",
    "VolumeSubdomain1D",
]



from importlib import metadata

try:
    __version__ = metadata.version("FESTIM")
except Exception:
    __version__ = "unknown"


R = 8.314462618  # Gas constant J.mol-1.K-1
k_B = 8.6173303e-5  # Boltzmann constant eV.K-1

from .advection import AdvectionTerm, VelocityField
from .boundary_conditions.dirichlet_bc import (
    DirichletBC,
    DirichletBCBase,
    FixedConcentrationBC,
    FixedTemperatureBC,
)
from .boundary_conditions.flux_bc import FluxBCBase, HeatFluxBC, ParticleFluxBC
from .boundary_conditions.henrys_bc import HenrysBC
from .boundary_conditions.sieverts_bc import SievertsBC
from .boundary_conditions.surface_reaction import SurfaceReactionBC
from .coupled_heat_hydrogen_problem import (
    CoupledTransientHeatTransferHydrogenTransport,
)
from .exports.average_surface import AverageSurface
from .exports.average_volume import AverageVolume
from .exports.maximum_surface import MaximumSurface
from .exports.maximum_volume import MaximumVolume
from .exports.minimum_surface import MinimumSurface
from .exports.minimum_volume import MinimumVolume
from .exports.profile_1d import Profile1DExport
from .exports.surface_flux import SurfaceFlux
from .exports.surface_quantity import SurfaceQuantity
from .exports.total_surface import TotalSurface
from .exports.total_volume import TotalVolume
from .exports.volume_quantity import VolumeQuantity
from .exports.vtx import ExportBaseClass, VTXSpeciesExport, VTXTemperatureExport
from .exports.xdmf import XDMFExport
from .heat_transfer_problem import HeatTransferProblem
from .helpers import (
    Value,
    as_fenics_constant,
    as_fenics_interp_expr_and_function,
    as_mapped_function,
    get_interpolation_points,
)
from .hydrogen_transport_problem import (
    HydrogenTransportProblem,
    HydrogenTransportProblemDiscontinuous,
    HydrogenTransportProblemDiscontinuousChangeVar,
    InterfaceMethod,
)
from .initial_condition import (
    InitialConcentration,
    InitialConditionBase,
    InitialTemperature,
    read_function_from_file,
)
from .material import Material, SolubilityLaw
from .mesh.mesh import Mesh
from .mesh.mesh_1d import Mesh1D
from .mesh.mesh_from_xdmf import MeshFromXDMF
from .problem import ProblemBase
from .reaction import Reaction
from .settings import Settings
from .source import HeatSource, ParticleSource, SourceBase
from .species import ImplicitSpecies, Species, find_species_from_name
from .stepsize import Stepsize
from .subdomain.interface import Interface
from .subdomain.surface_subdomain import (
    SurfaceSubdomain,
    SurfaceSubdomain1D,
    find_surface_from_id,
)
from .subdomain.volume_subdomain import (
    VolumeSubdomain,
    VolumeSubdomain1D,
    find_volume_from_id,
)
from .trap import Trap



from collections.abc import Callable

import basix
import ufl
from dolfinx import fem

from festim.helpers import Value, nmm_interpolate
from festim.species import Species
from festim.subdomain import VolumeSubdomain


class AdvectionTerm:
    """
    Advection term class

    args:
        velocity: the velocity field or function
        subdomain: the volume subdomain where the velocity is to be applied
        species: the species to which the velocity field is acting on

    attributes:
        velocity: the velocity field or function
        subdomain: the volume subdomain where the velocity is to be applied
        species: the species to which the velocity field is acting on

    """

    velocity: fem.Function
    subdomain: VolumeSubdomain
    species: Species

    def __init__(
        self,
        velocity: fem.Function,
        subdomain: VolumeSubdomain,
        species: Species,
    ):
        self.velocity = velocity
        self.subdomain = subdomain
        self.species = species

    @property
    def velocity(self):
        return self._velocity

    @velocity.setter
    def velocity(self, value):
        err_message = f"velocity must be a fem.Function, or callable not {type(value)}"
        if value is None:
            self._velocity = VelocityField(value)
        elif isinstance(
            value,
            fem.Function,
        ):
            self._velocity = VelocityField(value)
        elif isinstance(value, fem.Constant | fem.Expression | ufl.core.expr.Expr):
            raise TypeError(err_message)
        elif callable(value):
            self._velocity = VelocityField(value)
        else:
            raise TypeError(err_message)

    @property
    def subdomain(self):
        return self._subdomain

    @subdomain.setter
    def subdomain(self, value):
        if value is None:
            self._subdomain = value
        elif isinstance(value, VolumeSubdomain):
            self._subdomain = value
        else:
            raise TypeError(
                f"Subdomain must be a festim.Subdomain object, not {type(value)}"
            )

    @property
    def species(self) -> list[Species]:
        return self._species

    @species.setter
    def species(self, value):
        if not isinstance(value, list):
            value = [value]
        # check that all species are of type festim.Species
        for spe in value:
            if not isinstance(spe, Species):
                raise TypeError(
                    f"elements of species must be of type festim.Species not "
                    f"{type(spe)}"
                )
        self._species = value


class VelocityField(Value):
    """
    A class to handle input values of velocity fields from users and convert them to a
    relevent fenics object

    Args:
        input_value: The value of the user input

    Attributes:
        input_value : The value of the user input
        fenics_interpolation_expression : The expression of the user input that is used
            to update the `fenics_object`
        fenics_object : The value of the user input in fenics format
        explicit_time_dependent : True if the user input value is explicitly time
            dependent
        temperature_dependent : True if the user input value is temperature dependent
        vector_function_space: the vector function space of the fenics object
    """

    input_value: fem.Function | Callable

    fenics_object: fem.Function
    explicit_time_dependent: bool
    temperature_dependent: bool
    vector_function_space: fem.FunctionSpace

    def convert_input_value(
        self,
        function_space: fem.FunctionSpace,
        t: fem.Constant | None = None,
    ):
        """Converts a user given value to a relevent fenics object

        Args:
            function_space: the function space of the fenics object
            t: the time, optional
        """

        if isinstance(self.input_value, fem.Function):
            vel = self.input_value

        elif callable(self.input_value):
            # if callable function has args other than time, t, raise Typer Error
            args = self.input_value.__code__.co_varnames
            if args != ("t",):
                raise TypeError(
                    "velocity function can only be a function of time arg t"
                )

            vel = self.input_value(t)

            # if function does not return an fem.Fucntion, raise Typer Error
            if not isinstance(vel, fem.Function):
                raise ValueError(
                    "A time dependent advection field should return an fem.Function"
                    f", not a {type(vel)}"
                )

        # create vector function space and function
        v_cg = basix.ufl.element(
            "Lagrange",
            function_space.mesh.topology.cell_name(),
            1,
            shape=(function_space.mesh.topology.dim,),
        )
        self.vector_function_space = fem.functionspace(function_space.mesh, v_cg)
        self.fenics_object = fem.Function(self.vector_function_space)

        # interpolate input value into fenics object function
        nmm_interpolate(self.fenics_object, vel)

    def update(self, t: fem.Constant):
        """Updates the velocity field

        Args:
            t: the time
        """
        nmm_interpolate(self.fenics_object, self.input_value(t=float(t)))



import types
from collections.abc import Callable

import dolfinx
import numpy as np
from dolfinx.mesh import Mesh, locate_entities
from numpy import typing as npt
from scifem.mesh import transfer_meshtags_to_submesh

from festim.material import Material

# Define the appropriate method based on the version
try:
    from dolfinx.mesh import EntityMap

    entity_map_type = EntityMap
except ImportError:
    entity_map_type = npt.NDArray[np.int32]


class VolumeSubdomain:
    """
    Volume subdomain class

    Args:
        id (int): the id of the volume subdomain
        material (festim.Material): the material assigned to the subdomain
    """

    id: int
    submesh: dolfinx.mesh.Mesh
    cell_map: "entity_map_type"
    parent_mesh: dolfinx.mesh.Mesh
    parent_to_submesh: "entity_map_type"
    v_map: "entity_map_type"
    n_map: np.ndarray
    facet_to_parent: np.ndarray
    ft: dolfinx.mesh.MeshTags
    padded: bool  # NOTE: Once 0.9 support is dropped, this can be removed
    u: dolfinx.fem.Function
    u_n: dolfinx.fem.Function
    material: Material

    def __init__(self, id, material, locator: Callable | None = None):
        self.id = id
        self.material = material
        self.locator = locator

    def create_subdomain(self, mesh: dolfinx.mesh.Mesh, marker: dolfinx.mesh.MeshTags):
        """
        Creates the following attributes: ``.parent_mesh``, ``.submesh``, ``.cell_map``,
        ``.v_map``, ``padded``, and the entity map ``parent_to_submesh``.

        Only used in ``festim.HydrogenTransportProblemDiscontinuous``

        Args:
            mesh (dolfinx.mesh.Mesh): the parent mesh
            marker (dolfinx.mesh.MeshTags): the parent volume markers
        """
        assert marker.dim == mesh.topology.dim
        self.parent_mesh = (
            mesh  # NOTE: it doesn't seem like we use this attribute anywhere
        )
        entities = marker.find(self.id)
        self.submesh, self.cell_map, self.v_map, self.n_map = (
            dolfinx.mesh.create_submesh(mesh, marker.dim, entities)
        )
        if isinstance(entity_map_type, types.GenericAlias):
            num_cells_local = (
                mesh.topology.index_map(marker.dim).size_local
                + mesh.topology.index_map(marker.dim).num_ghosts
            )
            self.parent_to_submesh = np.full(num_cells_local, -1, dtype=np.int32)
            self.parent_to_submesh[self.cell_map] = np.arange(
                len(self.cell_map), dtype=np.int32
            )

            self.padded = False

    def transfer_meshtag(self, mesh: dolfinx.mesh.Mesh, tag: dolfinx.mesh.MeshTags):
        # Transfer meshtags to submesh
        assert self.submesh is not None, "Need to call create_subdomain first"
        self.ft, self.facet_to_parent = transfer_meshtags_to_submesh(
            tag, self.submesh, self.v_map, self.cell_map
        )

    def locate_subdomain_entities(self, mesh: Mesh) -> npt.NDArray[np.int32]:
        """Locates all cells in subdomain borders within domain

        Args:
            mesh: the mesh of the model

        Returns:
            entities: the entities of the subdomain
        """
        if self.locator is None:
            raise ValueError("No locator function provided for locating cells.")

        entities = locate_entities(mesh, mesh.topology.dim, self.locator)
        return entities


class VolumeSubdomain1D(VolumeSubdomain):
    """
    Volume subdomain class for 1D cases

    Args:
        id (int): the id of the volume subdomain
        borders (list of float): the borders of the volume subdomain
        material (festim.Material): the material of the volume subdomain

    Attributes:
        id (int): the id of the volume subdomain
        borders (list of float): the borders of the volume subdomain
        material (festim.Material): the material of the volume subdomain

    Examples:

        .. testsetup:: VolumeSubdomain1D

            from festim import VolumeSubdomain1D, Material
            my_mat = Material(D_0=1, E_D=1, name="test_mat")

        .. testcode:: VolumeSubdomain1D

            VolumeSubdomain1D(id=1, borders=[0, 1], material=my_mat)
    """

    def __init__(self, id, borders, material) -> None:
        super().__init__(
            id,
            material,
            locator=lambda x: np.logical_and(x[0] >= borders[0], x[0] <= borders[1]),
        )
        self.borders = borders


def find_volume_from_id(id: int, volumes: list):
    """Returns the correct volume subdomain object from a list of volume ids
    based on an int

    Args:
        id (int): the id of the volume subdomain
        volumes (list): the list of volumes

    Returns:
        festim.VolumeSubdomain: the volume subdomain object with the correct id

    Raises:
        ValueError: if the volume name is not found in the list of volumes

    """
    for vol in volumes:
        if vol.id == id:
            return vol
    raise ValueError(f"id {id} not found in list of volumes")



import numpy as np
import ufl
from dolfinx import fem

from festim.helpers import Value
from festim.species import Species
from festim.subdomain.volume_subdomain import VolumeSubdomain


class SourceBase:
    """
    Source base class

    Args:
        value: the value of the source
        volume: the volume subdomains where the source is applied

    Attributes:
        value: the value of the source
        volume: the volume subdomains where the source is applied
    """

    value: (
        float
        | int
        | fem.Constant
        | np.ndarray
        | fem.Expression
        | ufl.core.expr.Expr
        | fem.Function
    )
    volume: VolumeSubdomain

    def __init__(
        self,
        value: (
            float
            | int
            | fem.Constant
            | np.ndarray
            | fem.Expression
            | ufl.core.expr.Expr
            | fem.Function
            | Value
        ),
        volume: VolumeSubdomain,
    ):
        self.value = value
        self.volume = volume

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        if isinstance(value, Value):
            self._value = value
        else:
            self._value = Value(value)

    @property
    def volume(self):
        return self._volume

    @volume.setter
    def volume(self, value):
        # check that volume is festim.VolumeSubdomain
        if not isinstance(value, VolumeSubdomain):
            raise TypeError("volume must be of type festim.VolumeSubdomain")
        self._volume = value


class ParticleSource(SourceBase):
    """
    Particle source class

    Args:
        value: the value of the source
        volume: the volume subdomains where the source is applied
        species: the species to which the source is applied

    Attributes:
        value: the value of the source
        volume: the volume subdomains where the source is applied
        species: the species to which the source is applied

    Examples:

        .. highlight:: python
        .. code-block:: python

            from festim import ParticleSource

            ParticleSource(volume=my_vol, value=1, species="H")
            ParticleSource(volume=my_vol, value=lambda x: 1 + x[0], species="H")
            ParticleSource(volume=my_vol, value=lambda t: 1 + t, species="H")
            ParticleSource(volume=my_vol, value=lambda T: 1 + T, species="H")
            ParticleSource(volume=my_vol, value=lambda x, t: 1 + x[0] + t, species="H")
    """

    species: Species

    def __init__(self, value, volume, species: Species):
        self.species = species
        super().__init__(value, volume)

    @property
    def species(self):
        return self._species

    @species.setter
    def species(self, value):
        # check that species is festim.Species or list of festim.Species
        if not isinstance(value, Species):
            raise TypeError("species must be of type festim.Species")

        self._species = value


class HeatSource(SourceBase):
    """
    Heat source class

    Args:
        value: the value of the source
        volume: the volume subdomains where the source is applied

    Attributes:
        value: the value of the source
        volume: the volume subdomains where the source is applied

    Examples:

        .. highlight:: python
        .. code-block:: python

            from festim import HeatSource

            HeatSource(volume=my_vol, value=1)
            HeatSource(volume=my_vol, value=lambda x: 1 + x[0])
            HeatSource(volume=my_vol, value=lambda t: 1 + t)
            HeatSource(volume=my_vol, value=lambda x, t: 1 + x[0] + t)
    """

    def __init__(self, value, volume):
        super().__init__(value, volume)

        if self.value.temperature_dependent:
            raise ValueError("Heat source cannot be temperature dependent")



import mpi4py.MPI as MPI

import matplotlib.pyplot as plt
import numpy as np
import ufl
from dolfinx import fem

import festim as F


def source_from_exact_solution(
    exact_solution, thermal_conductivity, density, heat_capacity
):
    import sympy as sp
    from sympy.vector import CoordSys3D, divergence, gradient

    R = CoordSys3D("R")
    x = [R.x, R.y, R.z]
    t = sp.symbols("t")

    u = exact_solution(x, t)
    density = density(x, t)
    heat_capacity = heat_capacity(x, t)
    thermal_cond = thermal_conductivity(x, t)
    source = density * heat_capacity * sp.diff(u, t) - divergence(
        thermal_cond * gradient(u, x), x
    )

    return sp.lambdify([x, t], source)


def error_L2(u_computed, u_exact, degree_raise=3):
    # Create higher order function space
    degree = u_computed.function_space.ufl_element().degree()
    family = u_computed.function_space.ufl_element().family()
    mesh = u_computed.function_space.mesh
    W = fem.FunctionSpace(mesh, (family, degree + degree_raise))
    # Interpolate approximate solution
    u_W = fem.Function(W)
    u_W.interpolate(u_computed)

    # Interpolate exact solution, special handling if exact solution
    # is a ufl expression or a python lambda function
    u_ex_W = fem.Function(W)
    if isinstance(u_exact, ufl.core.expr.Expr):
        u_expr = fem.Expression(u_exact, W.element.interpolation_points)
        u_ex_W.interpolate(u_expr)
    else:
        u_ex_W.interpolate(u_exact)

    # Compute the error in the higher order function space
    e_W = fem.Function(W)
    e_W.x.array[:] = u_W.x.array - u_ex_W.x.array

    # Integrate the error
    error = fem.form(ufl.inner(e_W, e_W) * ufl.dx)
    error_local = fem.assemble_scalar(error)
    error_global = mesh.comm.allreduce(error_local, op=MPI.SUM)
    return np.sqrt(error_global)


def run(N):
    def exact_solution(x, t):
        return 2 * x[0] ** 2 + 20 * t

    def density(T):
        return 0.2 * T + 2

    def heat_capacity(T):
        return 0.2 * T + 3

    def thermal_conductivity(T):
        return 0.1 * T + 4

    mms_source_from_sp = source_from_exact_solution(
        exact_solution,
        density=lambda x, t: density(exact_solution(x, t)),
        heat_capacity=lambda x, t: heat_capacity(exact_solution(x, t)),
        thermal_conductivity=lambda x, t: thermal_conductivity(exact_solution(x, t)),
    )

    def mms_source(x, t):
        return mms_source_from_sp((x[0], None, None), t)

    my_problem = F.HeatTransferProblem()

    my_problem.mesh = F.Mesh1D(vertices=np.linspace(2, 3, N))
    left = F.SurfaceSubdomain1D(id=1, x=2)
    right = F.SurfaceSubdomain1D(id=2, x=3)
    my_problem.surface_subdomains = [left, right]
    mat = F.Material(D_0=None, E_D=None)
    mat.thermal_conductivity = thermal_conductivity
    mat.density = density
    mat.heat_capacity = heat_capacity

    my_problem.volume_subdomains = [
        F.VolumeSubdomain1D(id=1, borders=[2, 3], material=mat)
    ]

    # NOTE: it's good to check that without the IC the solution is not the exact one
    my_problem.initial_condition = F.InitialTemperature(lambda x: exact_solution(x, 0))

    my_problem.boundary_conditions = [
        F.FixedTemperatureBC(subdomain=left, value=exact_solution),
        F.FixedTemperatureBC(subdomain=right, value=exact_solution),
    ]

    my_problem.sources = [
        F.HeatSource(value=mms_source, volume=my_problem.volume_subdomains[0])
    ]

    my_problem.settings = F.Settings(
        atol=1e-8,
        rtol=1e-10,
        final_time=1,  # final time shouldn't be too long so that a potential error at the initial timestep is not negligible
    )

    # Forward euler isn't great so dt should be small
    # although it's ok here since the time derivative is constant
    my_problem.settings.stepsize = F.Stepsize(0.05)

    my_problem.exports = [
        F.VTXSpeciesExport(filename="test_transient_heat_transfer.bp")
    ]

    my_problem.initialise()
    my_problem.run()

    computed_solution = my_problem.u
    # we use the exact final time of the simulation which may differ from the one specified in the settings
    final_time_sim = my_problem.t.value

    def exact_solution_end(x):
        return exact_solution(x, final_time_sim)

    L2_error = error_L2(computed_solution, exact_solution_end)
    return L2_error


if __name__ == "__main__":
    Ns = np.geomspace(10, 3300, 10)
    errors = []
    for N in Ns:
        L2_error = run(int(N))
        print(f"With {int(N)} cells, L2_error = {L2_error}")
        errors.append(L2_error)

    plt.loglog(Ns, errors, "-o", label="L2")
    # plot a slope of 2
    plt.loglog(Ns, 1 / Ns**2, "--", color="black", label="order 2")
    plt.xlabel("Number of cells")
    plt.ylabel("Error")
    plt.legend()
    plt.show()



from .average_surface import AverageSurface
from .average_volume import AverageVolume
from .maximum_surface import MaximumSurface
from .maximum_volume import MaximumVolume
from .minimum_surface import MinimumSurface
from .minimum_volume import MinimumVolume
from .profile_1d import Profile1DExport
from .surface_flux import SurfaceFlux
from .surface_quantity import SurfaceQuantity
from .total_surface import TotalSurface
from .total_volume import TotalVolume
from .volume_quantity import VolumeQuantity
from .vtx import ExportBaseClass, VTXSpeciesExport, VTXTemperatureExport
from .xdmf import XDMFExport

__all__ = [
    "AverageSurface",
    "AverageVolume",
    "ExportBaseClass",
    "MaximumSurface",
    "MaximumVolume",
    "MinimumSurface",
    "MinimumVolume",
    "Profile1DExport",
    "SurfaceFlux",
    "SurfaceQuantity",
    "TotalSurface",
    "TotalVolume",
    "VTXSpeciesExport",
    "VTXTemperatureExport",
    "VolumeQuantity",
    "XDMFExport",
]



import numpy as np
import ufl
from dolfinx import fem

import festim as F

from .tools import error_L2

test_mesh_1d = F.Mesh1D(np.linspace(0, 1, 10000))
x_1d = ufl.SpatialCoordinate(test_mesh_1d.mesh)


def test_flux_bc_1_mobile_MMS_steady_state():
    """
    MMS test with a flux BC considering one mobile species at steady state
    """

    u_exact = lambda x: 1 + 2 * x[0] ** 2

    V = fem.functionspace(test_mesh_1d.mesh, ("Lagrange", 1))
    T = fem.Function(V)

    D_0 = 1
    E_D = 0.1
    T = 500
    D = D_0 * ufl.exp(-E_D / (F.k_B * T))

    my_model = F.HydrogenTransportProblem()
    my_model.mesh = test_mesh_1d
    my_mat = F.Material(name="mat", D_0=D_0, E_D=E_D)
    vol = F.VolumeSubdomain1D(id=1, borders=[0, 1], material=my_mat)
    left = F.SurfaceSubdomain1D(id=1, x=0)
    right = F.SurfaceSubdomain1D(id=2, x=1)

    my_model.subdomains = [vol, left, right]

    H = F.Species("H")
    my_model.species = [H]

    my_model.temperature = 500

    my_model.boundary_conditions = [
        F.DirichletBC(subdomain=left, value=u_exact, species=H),
        F.ParticleFluxBC(subdomain=right, value=4 * D, species=H),
    ]

    f = -ufl.div(D * ufl.grad(u_exact(x_1d)))
    my_model.sources = [F.ParticleSource(value=f, volume=vol, species=H)]

    my_model.settings = F.Settings(atol=1e-10, rtol=1e-10, transient=False)

    my_model.initialise()
    my_model.run()

    H_computed = H.post_processing_solution

    L2_error = error_L2(H_computed, u_exact)

    assert L2_error < 1e-7


def test_flux_bc_heat_transfer_steady_state():
    """
    MMS test with a flux BC in a heat transfer problem at steady state
    """

    u_exact = lambda x: 1 + 2 * x[0] ** 2

    thermal_cond = 2.3

    my_model = F.HeatTransferProblem()
    my_model.mesh = test_mesh_1d
    my_mat = F.Material(name="mat", D_0=1, E_D=1, thermal_conductivity=thermal_cond)
    vol = F.VolumeSubdomain1D(id=1, borders=[0, 1], material=my_mat)
    left = F.SurfaceSubdomain1D(id=1, x=0)
    right = F.SurfaceSubdomain1D(id=2, x=1)

    my_model.subdomains = [vol, left, right]

    my_model.boundary_conditions = [
        F.FixedTemperatureBC(subdomain=left, value=u_exact),
        F.HeatFluxBC(subdomain=right, value=4 * thermal_cond),
    ]

    f = -ufl.div(thermal_cond * ufl.grad(u_exact(x_1d)))
    my_model.sources = [F.HeatSource(value=f, volume=vol)]

    my_model.settings = F.Settings(atol=1e-10, rtol=1e-10, transient=False)

    my_model.initialise()
    my_model.run()

    T_computed = my_model.u

    L2_error = error_L2(T_computed, u_exact)

    assert L2_error < 1e-7
